"""
Module with useful time/date functions
"""

import datetime
import dateutil.parser

from .constants import SHConstants


def get_dates_in_range(start_date, end_date):
    """ Get all dates within input start and end date in ISO 8601 format

    :param start_date: start date in ISO 8601 format
    :type start_date: str
    :param end_date: end date in ISO 8601 format
    :type end_date: str
    :return: list of dates between start_date and end_date in ISO 8601 format
    :rtype: list of str
    """
    start_dt = iso_to_datetime(start_date)
    end_dt = iso_to_datetime(end_date)
    num_days = int((end_dt - start_dt).days)
    return [datetime_to_iso(start_dt + datetime.timedelta(i)) for i in range(num_days + 1)]


def next_date(date):
    """ Get date of day after input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-13'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date after input date in ISO 8601 format
    :rtype: str
    """
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm + datetime.timedelta(1))


def prev_date(date):
    """ Get date of day previous to input date in ISO 8601 format

    For instance, if input date is ``'2017-03-12'``, the function returns ``'2017-03-11'``

    :param date: input date in ISO 8601 format
    :type date: str
    :return: date previous to input date in ISO 8601 format
    :rtype: str
    """
    dtm = iso_to_datetime(date)
    return datetime_to_iso(dtm - datetime.timedelta(1))


def iso_to_datetime(date):
    """ Convert ISO 8601 time format to datetime format

    This function converts a date in ISO format, e.g. ``2017-09-14`` to a `datetime` instance, e.g.
    ``datetime.datetime(2017,9,14,0,0)``

    :param date: date in ISO 8601 format
    :type date: str
    :return: datetime instance
    :rtype: datetime
    """
    chunks = list(map(int, date.split('T')[0].split('-')))
    return datetime.datetime(chunks[0], chunks[1], chunks[2])


def datetime_to_iso(date, only_date=True):
    """ Convert datetime format to ISO 8601 time format

    This function converts a date in datetime instance, e.g. ``datetime.datetime(2017,9,14,0,0)`` to ISO format,
    e.g. ``2017-09-14``

    :param date: datetime instance to convert
    :type date: datetime
    :param only_date: whether to return date only or also time information. Default is `True`
    :type only_date: bool
    :return: date in ISO 8601 format
    :rtype: str
    """
    if only_date:
        return date.isoformat().split('T')[0]
    return date.isoformat()


def get_current_date():
    """ Get current date in ISO 8601 format

    :return: current date in ISO 8601 format
    :rtype: str
    """
    return datetime_to_iso(datetime.datetime.now())


def is_valid_time(time):
    """ Check if input string represents a valid time/date stamp

    :param time: a string containing a time/date stamp
    :type time: str
    :return: `True` is string is a valid time/date stamp, `False` otherwise
    :rtype: bool
    """
    try:
        dateutil.parser.parse(time)
        return True
    except BaseException:
        return False


def parse_time(time_input):
    """ Parse input time/date string into ISO 8601 string

    :param time_input: time/date to parse
    :type time_input: str or datetime.date or datetime.datetime
    :return: parsed string in ISO 8601 format
    :rtype: str
    """
    if isinstance(time_input, datetime.date):
        return time_input.isoformat()  # datetime.date only returns date, datetime.datetime also returns time

    if len(time_input) < 8:
        raise ValueError('Invalid time string {}.\n'
                         'Please specify time in formats YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS'.format(time_input))
    time = dateutil.parser.parse(time_input)
    if len(time_input) <= 10:
        return time.date().isoformat()
    return time.isoformat()


def parse_time_interval(time):
    """ Parse input into an interval of two times, specifying start and end time, in ISO 8601 format, for example:

    ``(2017-01-15:T00:00:00, 2017-01-16:T23:59:59)``

    The input time can have the following formats, which will be parsed as:

    * `YYYY-MM-DD` -> `[YYYY-MM-DD:T00:00:00, YYYY-MM-DD:T23:59:59]`
    * `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`
    * list or tuple of two dates in form `YYYY-MM-DD` -> `[YYYY-MM-DDT00:00:00, YYYY-MM-DDT23:59:59]`
    * list or tuple of two dates in form `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`

    All input times can also be specified as `datetime` objects. Instances of `datetime.date` will be treated as
    `YYYY-MM-DD` and instance of `datetime.datetime` will be treated as `YYYY-MM-DDThh:mm:ss`.

    :param time: An input time
    :type time: str or datetime.datetime
    :return: interval of start and end date of the form `YYYY-MM-DDThh:mm:ss`
    :rtype: (str, str)
    :raises: ValueError
    """
    if time is SHConstants.LATEST:
        date_interval = '1985-01-01', get_current_date()
    elif isinstance(time, (str, datetime.date)):
        parsed_time = parse_time(time)
        date_interval = parsed_time, parsed_time
    elif isinstance(time, (tuple, list)) and len(time) == 2:
        date_interval = parse_time(time[0]), parse_time(time[1])
    else:
        raise ValueError('Time must be a string/datetime object or tuple/list of 2 strings/datetime objects')

    if 'T' not in date_interval[0]:
        date_interval = date_interval[0] + 'T00:00:00', date_interval[1]
    if 'T' not in date_interval[1]:
        date_interval = date_interval[0], date_interval[1] + 'T23:59:59'

    if date_interval[1] < date_interval[0]:
        raise ValueError('Start of time interval is larger than end of time interval')

    return date_interval



"""
Module for data decoding
"""
import json
import struct
import tarfile
import warnings
from io import BytesIO
from xml.etree import ElementTree

import numpy as np
import tifffile as tiff
from PIL import Image

from .constants import MimeType
from .exceptions import ImageDecodingError


warnings.simplefilter('ignore', Image.DecompressionBombWarning)


def decode_data(response_content, data_type):
    """ Interprets downloaded data and returns it.

    :param response_content: downloaded data (i.e. json, png, tiff, xml, zip, ... file)
    :type response_content: bytes
    :param data_type: expected downloaded data type
    :type data_type: constants.MimeType
    :return: downloaded data
    :rtype: numpy array in case of image data type, or other possible data type
    :raises: ValueError
    """
    if data_type is MimeType.JSON:
        return json.loads(response_content.decode('utf-8'))
    if data_type is MimeType.TAR:
        return decode_tar(response_content)
    if MimeType.is_image_format(data_type):
        return decode_image(response_content, data_type)
    if data_type is MimeType.XML or data_type is MimeType.GML or data_type is MimeType.SAFE:
        return ElementTree.fromstring(response_content)

    try:
        return {
            MimeType.TAR: decode_tar,
            MimeType.RAW: response_content,
            MimeType.TXT: response_content,
            MimeType.ZIP: BytesIO(response_content)
        }[data_type]
    except KeyError:
        raise ValueError('Unknown response data type {}'.format(data_type))


def decode_image(data, image_type):
    """ Decodes the image provided in various formats, i.e. png, 16-bit float tiff, 32-bit float tiff, jp2
    and returns it as an numpy array

    :param data: image in its original format
    :type data: any of possible image types
    :param image_type: expected image format
    :type image_type: constants.MimeType
    :return: image as numpy array
    :rtype: numpy array
    :raises: ImageDecodingError
    """
    bytes_data = BytesIO(data)
    if image_type.is_tiff_format():
        image = tiff.imread(bytes_data)
    else:
        image = np.array(Image.open(bytes_data))

        if image_type is MimeType.JP2:
            try:
                bit_depth = get_jp2_bit_depth(bytes_data)
                image = fix_jp2_image(image, bit_depth)
            except ValueError:
                pass

    if image is None:
        raise ImageDecodingError('Unable to decode image')
    return image


def decode_tar(data):
    """ A decoder to convert response bytes into a dictionary of {filename: value}

    :param data: Data to decode
    :type data: bytes
    :return: A dictionary of decoded files from a tar file
    :rtype: dict(str: object)
    """
    tar = tarfile.open(fileobj=BytesIO(data))
    itr = ((member.name, get_data_format(member.name), tar.extractfile(member)) for member in tar.getmembers())

    return {filename: decode_data(file.read(), file_type) for filename, file_type, file in itr}


def decode_sentinelhub_err_msg(response):
    """ Decodes error message from Sentinel Hub service

    :param response: Sentinel Hub service response
    :type response: requests.Response
    :return: An error message
    :rtype: str
    """
    try:
        server_message = []
        for elem in decode_data(response.content, MimeType.XML):
            if 'ServiceException' in elem.tag or 'Message' in elem.tag:
                server_message.append(elem.text.strip('\n\t '))
        return ''.join(server_message)
    except ElementTree.ParseError:
        return response.text


def get_jp2_bit_depth(stream):
    """ Reads bit encoding depth of jpeg2000 file in binary stream format

    :param stream: binary stream format
    :type stream: Binary I/O (e.g. io.BytesIO, io.BufferedReader, ...)
    :return: bit depth
    :rtype: int
    """
    stream.seek(0)
    while True:
        read_buffer = stream.read(8)
        if len(read_buffer) < 8:
            raise ValueError('Image Header Box not found in Jpeg2000 file')

        _, box_id = struct.unpack('>I4s', read_buffer)

        if box_id == b'ihdr':
            read_buffer = stream.read(14)
            params = struct.unpack('>IIHBBBB', read_buffer)
            return (params[3] & 0x7f) + 1


def fix_jp2_image(image, bit_depth):
    """ Because Pillow library incorrectly reads JPEG 2000 images with 15-bit encoding this function corrects the
    values in image.

    :param image: image read by opencv library
    :type image: numpy array
    :param bit_depth: bit depth of jp2 image encoding
    :type bit_depth: int
    :return: corrected image
    :rtype: numpy array
    """
    if bit_depth in [8, 16]:
        return image
    if bit_depth == 15:
        try:
            return image >> 1
        except TypeError:
            raise IOError('Failed to read JPEG 2000 image correctly. Most likely reason is that Pillow did not '
                          'install OpenJPEG library correctly. Try reinstalling Pillow from a wheel')

    raise ValueError('Bit depth {} of jp2 image is currently not supported. '
                     'Please raise an issue on package Github page'.format(bit_depth))


def get_data_format(filename):
    """ Util function to guess format from filename extension

    :param filename: name of file
    :type filename: str
    :return: file extension
    :rtype: MimeType
    """
    fmt_ext = filename.split('.')[-1]
    return MimeType(MimeType.canonical_extension(fmt_ext))



""" Caching logic
"""
import hashlib
import json


def hash_request(url, payload):
    """ Joins url and payload into a dictionary, dumps it as json and hashes it with md5 algorithm
    """
    if not isinstance(url, str):
        raise ValueError('Expected the url argument of type str, got {}.'.format(type(url)))
    if not isinstance(payload, dict):
        raise ValueError('Expected the payload argument of type dict, got {}.'.format(type(payload)))

    hashable = dict(url=url, payload=payload)
    hashable = json.dumps(hashable, indent=4)
    hashed = hashlib.md5(hashable.encode('utf-8')).hexdigest()

    return hashed, hashable



"""
This module lists all externally useful classes and functions
"""

from .data_request import WmsRequest, WcsRequest, FisRequest, GeopediaWmsRequest, GeopediaImageRequest, \
    AwsTileRequest, AwsProductRequest, get_safe_format, download_safe_format

from .aws import AwsProduct, AwsTile
from .aws_safe import SafeProduct, SafeTile

from .areas import BBoxSplitter, OsmSplitter, TileSplitter, CustomGridSplitter, UtmGridSplitter, UtmZoneSplitter

from .ogc import WebFeatureService
from .geopedia import GeopediaFeatureIterator, GeopediaSession

from .geometry import BBox, Geometry, BBoxCollection
from .constants import DataSource, CustomUrlParam, CRS, MimeType, SHConstants, AwsConstants, ServiceType, \
    HistogramType

from .config import SHConfig

from .download import DownloadRequest, get_json, get_xml

from .exceptions import DownloadFailedException, AwsDownloadFailedException

from .opensearch import get_tile_info_id, get_tile_info, get_area_dates, get_area_info

from .io_utils import read_data, write_data
from .os_utils import get_content_list, get_folder_list, get_file_list, make_folder, create_parent_folder, rename, size
from .geo_utils import bbox_to_dimensions, bbox_to_resolution, get_image_dimension, to_utm_bbox, get_utm_bbox,\
    wgs84_to_utm, to_wgs84, utm_to_pixel, pixel_to_utm, wgs84_to_pixel, get_utm_crs, transform_point, transform_bbox
from .time_utils import next_date, prev_date, get_current_date

from .testing_utils import TestSentinelHub, TestCaseContainer

from .download.sentinelhub_client import SentinelHubDownloadClient
from .sentinelhub_session import SentinelHubSession

from .sentinelhub_request import body, bounds, data, output, response
from .download.cache import hash_request

from .time_utils import parse_time_interval

from ._version import __version__



"""
Module for working with Sentinel Hub OGC services
"""

import logging
import datetime
from base64 import b64encode
from urllib.parse import urlencode

import shapely.geometry
import shapely.wkt
import shapely.ops

from .constants import ServiceType, DataSource, MimeType, CRS, SHConstants, CustomUrlParam
from .config import SHConfig
from .geo_utils import get_image_dimension
from .geometry import BBox, Geometry
from .download import DownloadRequest, get_json
from .time_utils import parse_time_interval

LOGGER = logging.getLogger(__name__)


class OgcService:
    """ The base class for Sentinel Hub OGC services
    """
    def __init__(self, base_url=None, instance_id=None):
        """
        :param base_url: base url of Sentinel Hub's OGC services. If `None`, the url specified in the configuration
                    file is taken.
        :type base_url: str or None
        :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If `None`, the instance
            ID specified in the configuration file is taken.
        :type instance_id: str or None
        """
        self.base_url = SHConfig().get_sh_ogc_url() if not base_url else base_url
        self.instance_id = SHConfig().instance_id if not instance_id else instance_id

        if not self.instance_id:
            raise ValueError('Instance ID is not set. '
                             'Set it either in request initialization or in configuration file. '
                             'Check http://sentinelhub-py.readthedocs.io/en/latest/configure.html for more info.')

    @staticmethod
    def _filter_dates(dates, time_difference):
        """ Filters out dates within time_difference, preserving only the oldest date.

        :param dates: a list of datetime objects
        :param time_difference: a ``datetime.timedelta`` representing the time difference threshold
        :return: an ordered list of datetimes `d1<=d2<=...<=dn` such that `d[i+1]-di > time_difference`
        :rtype: list(datetime.datetime)
        """

        LOGGER.debug("dates=%s", dates)

        if len(dates) <= 1:
            return dates

        sorted_dates = sorted(dates)

        separate_dates = [sorted_dates[0]]
        for curr_date in sorted_dates[1:]:
            if curr_date - separate_dates[-1] > time_difference:
                separate_dates.append(curr_date)
        return separate_dates

    @staticmethod
    def _sentinel1_product_check(product_id, data_source):
        """ Checks if Sentinel-1 product ID matches Sentinel-1 DataSource configuration

        :param product_id: Sentinel-1 product ID
        :type product_id: str
        :param data_source: One of the supported Sentinel-1 data sources
        :type data_source: constants.DataSource
        :return: `True` if data_source contains product_id and `False` otherwise
        :rtype: bool
        """
        props = product_id.split('_')
        acquisition, resolution, polarisation = props[1], props[2][3], props[3][2:4]
        if acquisition in ['IW', 'EW'] and resolution in ['M', 'H'] and polarisation in ['DV', 'DH', 'SV', 'SH']:
            return acquisition == data_source.value[2].name and polarisation == data_source.value[3].name and \
                   resolution == data_source.value[4].name[0]
        raise ValueError('Unknown Sentinel-1 tile type: {}'.format(product_id))


class OgcImageService(OgcService):
    """Sentinel Hub OGC services class for providing image data

    Intermediate layer between QGC-type requests (WmsRequest and WcsRequest) and the Sentinel Hub OGC (WMS and WCS)
    services.
    """
    def __init__(self, **kwargs):
        """
        :param base_url: base url of Sentinel Hub's OGC services. If `None`, the url specified in the configuration
                    file is taken.
        :type base_url: str or None
        :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If `None`, the instance
            ID specified in the configuration file is taken.
        :type instance_id: str or None
        """
        super().__init__(**kwargs)

        self.wfs_iterator = None

    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: QGC-type request with specified bounding box, time interval, and cloud coverage for specific
                        product.
        :type request: OgcRequest or GeopediaRequest
        :return: list of DownloadRequests
        """
        size_x, size_y = self.get_image_dimensions(request)
        return [DownloadRequest(url=self.get_url(request=request, date=date, size_x=size_x, size_y=size_y),
                                filename=self.get_filename(request, date, size_x, size_y),
                                data_type=request.image_format, headers=SHConstants.HEADERS)
                for date in self.get_dates(request)]

    def get_url(self, request, *, date=None, size_x=None, size_y=None, geometry=None):
        """ Returns url to Sentinel Hub's OGC service for the product specified by the OgcRequest and date.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :type geometry: list of BBox or Geometry
        :return:  dictionary with parameters
        :return: url to Sentinel Hub's OGC service for this product.
        :rtype: str
        """
        url = self.get_base_url(request)
        authority = self.instance_id if hasattr(self, 'instance_id') else request.theme

        params = self._get_common_url_parameters(request)
        if request.service_type in (ServiceType.WMS, ServiceType.WCS):
            params = {**params, **self._get_wms_wcs_url_parameters(request, date)}
        if request.service_type is ServiceType.WMS:
            params = {**params, **self._get_wms_url_parameters(request, size_x, size_y)}
        elif request.service_type is ServiceType.WCS:
            params = {**params, **self._get_wcs_url_parameters(request, size_x, size_y)}
        elif request.service_type is ServiceType.FIS:
            params = {**params, **self._get_fis_parameters(request, geometry)}

        return '{}/{}?{}'.format(url, authority, urlencode(params))

    def get_base_url(self, request):
        """ Creates base url string.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :return: base string for url to Sentinel Hub's OGC service for this product.
        :rtype: str
        """
        url = '{}/{}'.format(self.base_url, request.service_type.value)
        # These 2 lines are temporal and will be removed after the use of uswest url wont be required anymore:
        if hasattr(request, 'data_source') and request.data_source.is_uswest_source():
            url = 'https://services-uswest2.sentinel-hub.com/ogc/{}'.format(request.service_type.value)

        if hasattr(request, 'data_source') and request.data_source not in DataSource.get_available_sources():
            raise ValueError("{} is not available for service at service available at {}. Try"
                             "changing 'sh_base_url' parameter in package configuration "
                             "file".format(request.data_source, SHConfig().get_sh_ogc_url()))
        return url

    @staticmethod
    def _get_common_url_parameters(request):
        """ Returns parameters common dictionary for WMS, WCS and FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :return:  dictionary with parameters
        :rtype: dict
        """
        params = {
            'SERVICE': request.service_type.value
        }

        if hasattr(request, 'maxcc'):
            params['MAXCC'] = 100.0 * request.maxcc

        if hasattr(request, 'custom_url_params') and request.custom_url_params is not None:
            params = {**params,
                      **{k.value: str(v) for k, v in request.custom_url_params.items()}}

            if CustomUrlParam.EVALSCRIPT.value in params:
                evalscript = params[CustomUrlParam.EVALSCRIPT.value]
                params[CustomUrlParam.EVALSCRIPT.value] = b64encode(evalscript.encode()).decode()

            if CustomUrlParam.GEOMETRY.value in params:
                geometry = params[CustomUrlParam.GEOMETRY.value]
                crs = request.bbox.crs

                if isinstance(geometry, Geometry):
                    if geometry.crs is not crs:
                        raise ValueError('Geometry object in custom_url_params should have the same CRS as given BBox')
                else:
                    geometry = Geometry(geometry, crs)

                if geometry.crs is CRS.WGS84:
                    geometry = geometry.reverse()

                params[CustomUrlParam.GEOMETRY.value] = geometry.wkt

        return params

    @staticmethod
    def _get_wms_wcs_url_parameters(request, date):
        """ Returns parameters common dictionary for WMS and WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :return:  dictionary with parameters
        :rtype: dict
        """
        params = {
            'BBOX': str(request.bbox.reverse()) if request.bbox.crs is CRS.WGS84 else str(request.bbox),
            'FORMAT': MimeType.get_string(request.image_format),
            'CRS': CRS.ogc_string(request.bbox.crs),
        }

        if date is not None:
            start_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date - request.time_difference
            end_date = date if request.time_difference < datetime.timedelta(
                seconds=0) else date + request.time_difference
            params['TIME'] = '{}/{}'.format(start_date.isoformat(), end_date.isoformat())

        return params

    @staticmethod
    def _get_wms_url_parameters(request, size_x, size_y):
        """ Returns parameters dictionary for WMS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return:  dictionary with parameters
        :rtype: dict
        """
        return {
            'WIDTH': size_x,
            'HEIGHT': size_y,
            'LAYERS': request.layer,
            'REQUEST': 'GetMap',
            'VERSION': '1.3.0'
        }

    @staticmethod
    def _get_wcs_url_parameters(request, size_x, size_y):
        """ Returns parameters dictionary for WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return:  dictionary with parameters
        :rtype: dict
        """
        return {
            'RESX': size_x,
            'RESY': size_y,
            'COVERAGE': request.layer,
            'REQUEST': 'GetCoverage',
            'VERSION': '1.1.2'
        }

    @staticmethod
    def _get_fis_parameters(request, geometry):
        """ Returns parameters dictionary for FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param geometry: list of bounding boxes or geometries
        :type geometry: list of BBox or Geometry
        :return:  dictionary with parameters
        :rtype: dict
        """
        date_interval = parse_time_interval(request.time)

        params = {
            'CRS': CRS.ogc_string(geometry.crs),
            'LAYER': request.layer,
            'RESOLUTION': request.resolution,
            'TIME': '{}/{}'.format(date_interval[0], date_interval[1])
        }

        if not isinstance(geometry, (BBox, Geometry)):
            raise ValueError('Each geometry must be an instance of sentinelhub.{} or sentinelhub.{} but {} '
                             'found'.format(BBox.__name__, Geometry.__name__, geometry))
        if geometry.crs is CRS.WGS84:
            geometry = geometry.reverse()
        if isinstance(geometry, Geometry):
            params['GEOMETRY'] = geometry.wkt
        else:
            params['BBOX'] = str(geometry)

        if request.bins:
            params['BINS'] = request.bins

        if request.histogram_type:
            params['TYPE'] = request.histogram_type.value

        return params

    @staticmethod
    def get_filename(request, date, size_x, size_y):
        """ Get filename location

        Returns the filename's location on disk where data is or is going to be stored.
        The files are stored in the folder specified by the user when initialising OGC-type
        of request. The name of the file has the following structure:

        {service_type}_{layer}_{crs}_{bbox}_{time}_{size_x}X{size_y}_*{custom_url_params}.{image_format}

        In case of `TIFF_d32f` a `'_tiff_depth32f'` is added at the end of the filename (before format suffix)
        to differentiate it from 16-bit float tiff.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :param date: acquisition date or None
        :type date: datetime.datetime or None
        :param size_x: horizontal image dimension
        :type size_x: int or str
        :param size_y: vertical image dimension
        :type size_y: int or str
        :return: filename for this request and date
        :rtype: str
        """
        filename = '_'.join([
            str(request.service_type.value),
            request.layer,
            str(request.bbox.crs),
            str(request.bbox).replace(',', '_'),
            '' if date is None else date.strftime("%Y-%m-%dT%H-%M-%S"),
            '{}X{}'.format(size_x, size_y)
        ])

        filename = OgcImageService.filename_add_custom_url_params(filename, request)

        return OgcImageService.finalize_filename(filename, request.image_format)

    @staticmethod
    def filename_add_custom_url_params(filename, request):
        """ Adds custom url parameters to filename string

        :param filename: Initial filename
        :type filename: str
        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :type request: OgcRequest or GeopediaRequest
        :return: Filename with custom url parameters in the name
        :rtype: str
        """
        if hasattr(request, 'custom_url_params') and request.custom_url_params is not None:
            for param, value in sorted(request.custom_url_params.items(),
                                       key=lambda parameter_item: parameter_item[0].value):
                filename = '_'.join([filename, param.value, str(value)])

        return filename

    @staticmethod
    def finalize_filename(filename, file_format=None):
        """ Replaces invalid characters in filename string, adds image extension and reduces filename length

        :param filename: Incomplete filename string
        :type filename: str
        :param file_format: Format which will be used for filename extension
        :type file_format: MimeType
        :return: Final filename string
        :rtype: str
        """
        for char in [' ', '/', '\\', '|', ';', ':', '\n', '\t']:
            filename = filename.replace(char, '')

        if file_format:
            suffix = str(file_format.value)
            if file_format.is_tiff_format() and file_format is not MimeType.TIFF:
                suffix = str(MimeType.TIFF.value)
                filename = '_'.join([filename, str(file_format.value).replace(';', '_')])

            filename = '.'.join([filename[:254 - len(suffix)], suffix])

        LOGGER.debug("filename=%s", filename)

        return filename  # Even in UNIX systems filename must have at most 255 bytes

    def get_dates(self, request):
        """ Get available Sentinel-2 acquisitions at least time_difference apart

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        When a time_difference threshold is set to a positive value, the function filters out all datetimes which
        are within the time difference. The oldest datetime is preserved, all others all deleted.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: List of dates of existing acquisitions for the given request
        :rtype: list(datetime.datetime) or [None]
        """
        if DataSource.is_timeless(request.data_source):
            return [None]

        date_interval = parse_time_interval(request.time)

        LOGGER.debug('date_interval=%s', date_interval)

        if request.wfs_iterator is None:
            self.wfs_iterator = WebFeatureService(request.bbox, date_interval, data_source=request.data_source,
                                                  maxcc=request.maxcc, base_url=self.base_url,
                                                  instance_id=self.instance_id)
        else:
            self.wfs_iterator = request.wfs_iterator

        dates = sorted(set(self.wfs_iterator.get_dates()))

        if request.time is SHConstants.LATEST:
            dates = dates[-1:]
        return OgcService._filter_dates(dates, request.time_difference)

    @staticmethod
    def get_image_dimensions(request):
        """ Verifies or calculates image dimensions.

        :param request: OGC-type request
        :type request: WmsRequest or WcsRequest
        :return: horizontal and vertical dimensions of requested image
        :rtype: (int or str, int or str)
        """
        if request.service_type is ServiceType.WCS or (isinstance(request.size_x, int) and
                                                       isinstance(request.size_y, int)):
            return request.size_x, request.size_y
        if not isinstance(request.size_x, int) and not isinstance(request.size_y, int):
            raise ValueError("At least one of parameters 'width' and 'height' must have an integer value")
        missing_dimension = get_image_dimension(request.bbox, width=request.size_x, height=request.size_y)
        if request.size_x is None:
            return missing_dimension, request.size_y
        if request.size_y is None:
            return request.size_x, missing_dimension
        raise ValueError("Parameters 'width' and 'height' must be integers or None")

    def get_wfs_iterator(self):
        """ Returns iterator over info about all satellite tiles used for the request

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
                 DataSource.DEM it returns None.
        :rtype: Iterator[dict] or None
        """
        return self.wfs_iterator


class WebFeatureService(OgcService):
    """ Class for interaction with Sentinel Hub WFS service

    The class is an iterator over info data of all available satellite tiles for requested parameters. It collects data
    from Sentinel Hub service only during the first iteration. During next iterations it returns already obtained data.
    The data is in the order returned by Sentinel Hub WFS service.
    """
    def __init__(self, bbox, time_interval, *, data_source=DataSource.SENTINEL2_L1C, maxcc=1.0, **kwargs):
        """
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :type bbox: geometry.BBox
        :param time_interval: interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD
        :type time_interval: (str, str)
        :param data_source: Source of requested satellite data. Default is Sentinel-2 L1C data.
        :type data_source: constants.DataSource
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :type maxcc: float
        :param base_url: base url of Sentinel Hub's OGC services. If `None`, the url specified in the configuration
                        file is taken.
        :type base_url: str or None
        :param instance_id: user's instance id granting access to Sentinel Hub's OGC services. If `None`, the instance
            ID specified in the configuration file is taken.
        :type instance_id: str or None
        """
        super().__init__(**kwargs)

        self.bbox = bbox
        self.time_interval = parse_time_interval(time_interval)
        self.data_source = data_source
        self.maxcc = maxcc

        self.tile_list = []
        self.index = 0
        self.feature_offset = 0

    def __iter__(self):
        """ Iteration method

        :return: iterator of dictionaries containing info about product tiles
        :rtype: Iterator[dict]
        """
        self.index = 0
        return self

    def __next__(self):
        """ Next method

        :return: dictionary containing info about product tiles
        :rtype: dict
        """
        while self.index >= len(self.tile_list) and self.feature_offset is not None:
            self._fetch_features()

        if self.index < len(self.tile_list):
            self.index += 1
            return self.tile_list[self.index - 1]

        raise StopIteration

    def _fetch_features(self):
        """ Collects data from WFS service

        :return: dictionary containing info about product tiles
        :rtype: dict
        """
        if self.feature_offset is None:
            return

        main_url = '{}/{}/{}?'.format(self.base_url, ServiceType.WFS.value, self.instance_id)

        params = {'SERVICE': ServiceType.WFS.value,
                  'REQUEST': 'GetFeature',
                  'TYPENAMES': DataSource.get_wfs_typename(self.data_source),
                  'BBOX': str(self.bbox.reverse()) if self.bbox.crs is CRS.WGS84 else str(self.bbox),
                  'OUTPUTFORMAT': MimeType.get_string(MimeType.JSON),
                  'SRSNAME': CRS.ogc_string(self.bbox.crs),
                  'TIME': '{}/{}'.format(self.time_interval[0], self.time_interval[1]),
                  'MAXCC': 100.0 * self.maxcc,
                  'MAXFEATURES': SHConfig().max_wfs_records_per_query,
                  'FEATURE_OFFSET': self.feature_offset}

        url = main_url + urlencode(params)

        LOGGER.debug("URL=%s", url)
        response = get_json(url)

        is_sentinel1 = self.data_source.is_sentinel1()
        for tile_info in response["features"]:
            if is_sentinel1:
                if self._sentinel1_product_check(tile_info['properties']['id'], self.data_source) and \
                        self.data_source.contains_orbit_direction(tile_info['properties']['orbitDirection']):
                    self.tile_list.append(tile_info)
            else:
                self.tile_list.append(tile_info)

        if len(response["features"]) < SHConfig().max_wfs_records_per_query:
            self.feature_offset = None
        else:
            self.feature_offset += SHConfig().max_wfs_records_per_query

    def get_dates(self):
        """ Returns a list of acquisition times from tile info data

        :return: List of acquisition times in the order returned by WFS service.
        :rtype: list(datetime.datetime)
        """
        tile_dates = []

        for tile_info in self:
            if not tile_info['properties']['date']:  # could be True for custom (BYOC) data sources
                tile_dates.append(None)
            else:
                tile_dates.append(datetime.datetime.strptime('{}T{}'.
                                                             format(tile_info['properties']['date'],
                                                                    tile_info['properties']['time'].split('.')[0]),
                                                             '%Y-%m-%dT%H:%M:%S'))

        return tile_dates

    def get_geometries(self):
        """ Returns a list of geometries from tile info data

        :return: List of multipolygon geometries in the order returned by WFS service.
        :rtype: list(shapely.geometry.MultiPolygon)
        """
        return [shapely.geometry.shape(tile_info['geometry']) for tile_info in self]

    def get_tiles(self):
        """ Returns list of tiles with tile name, date and AWS index

        :return: List of tiles in form of (tile_name, date, aws_index)
        :rtype: list((str, str, int))
        """
        return [self._parse_tile_url(tile_info['properties']['path']) for tile_info in self]

    @staticmethod
    def _parse_tile_url(tile_url):
        """ Extracts tile name, data and AWS index from tile URL

        :param tile_url: Location of tile at AWS
        :type tile_url: str
        :return: Tuple in a form (tile_name, date, aws_index)
        :rtype: (str, str, int)
        """
        props = tile_url.rsplit('/', 7)
        return ''.join(props[1:4]), '-'.join(props[4:7]), int(props[7])



"""
Module for managing configuration data from `config.json`
"""

import os
import json
from collections import OrderedDict


class SHConfig:
    """ A sentinelhub-py package configuration class.

    The class reads during its first initialization the configurable settings from
    ``./config.json`` file:

        - `instance_id`: An instance ID for Sentinel Hub service used for OGC requests.
        - `sh_client_id`: User's OAuth client ID for Sentinel Hub service
        - `sh_client_secret`: User's OAuth client secret for Sentinel Hub service
        - `sh_base_url`: There exist multiple deployed instances of Sentinel Hub service, this parameter defines the
            location of a specific service instance.
        - `geopedia_wms_url`: Base url for Geopedia WMS services.
        - `geopedia_rest_url`: Base url for Geopedia REST services.
        - `aws_access_key_id`: Access key for AWS Requester Pays buckets.
        - `aws_secret_access_key`: Secret access key for AWS Requester Pays buckets.
        - `aws_metadata_url`: Base url for publicly available metadata files
        - `aws_s3_l1c_bucket`: Name of Sentinel-2 L1C bucket at AWS s3 service.
        - `aws_s3_l2a_bucket`: Name of Sentinel-2 L2A bucket at AWS s3 service.
        - `opensearch_url`: Base url for Sentinelhub Opensearch service.
        - `max_wfs_records_per_query`: Maximum number of records returned for each WFS query.
        - `max_opensearch_records_per_query`: Maximum number of records returned for each Opensearch query.
        - `max_download_attempts`: Maximum number of download attempts from a single URL until an error will be raised.
        - `download_sleep_time`: Number of seconds between the failed download attempt and the next attempt.
        - `download_timeout_seconds`: Maximum number of seconds before download attempt is canceled.

    Usage in the code:

        * ``SHConfig().sh_base_url``
        * ``SHConfig().instance_id``

    """
    class _SHConfig:
        """ Internal class holding configuration parameters
        """
        CONFIG_PARAMS = OrderedDict([
            ('instance_id', ''),
            ('sh_client_id', ''),
            ('sh_client_secret', ''),
            ('sh_base_url', 'https://services.sentinel-hub.com'),
            ('geopedia_wms_url', 'http://service.geopedia.world'),
            ('geopedia_rest_url', 'https://www.geopedia.world/rest'),
            ('aws_access_key_id', ''),
            ('aws_secret_access_key', ''),
            ('aws_metadata_url', 'https://roda.sentinel-hub.com'),
            ('aws_s3_l1c_bucket', 'sentinel-s2-l1c'),
            ('aws_s3_l2a_bucket', 'sentinel-s2-l2a'),
            ('opensearch_url', 'http://opensearch.sentinel-hub.com/resto/api/collections/Sentinel2'),
            ('max_wfs_records_per_query', 100),
            ('max_opensearch_records_per_query', 500),
            ('max_download_attempts', 4),
            ('download_sleep_time', 5),
            ('download_timeout_seconds', 120)
        ])

        def __init__(self):
            self.instance_id = ''
            self.load_configuration()

        def _parse_configuration(self, config):
            """ Checks if configuration file contains all keys and parses values
            """
            for param in self.CONFIG_PARAMS:
                if param not in config:
                    raise ValueError("Configuration file does not contain '%s' parameter." % param)

            for param, default_param in self.CONFIG_PARAMS.items():
                param_type = type(default_param)
                if not isinstance(config[param], param_type):
                    raise ValueError("Value of parameter '{}' must be of type {}".format(param, param_type.__name__))

            if config['max_wfs_records_per_query'] > 100:
                raise ValueError("Value of config parameter 'max_wfs_records_per_query' must be at most 100")
            if config['max_opensearch_records_per_query'] > 500:
                raise ValueError("Value of config parameter 'max_opensearch_records_per_query' must be at most 500")

            # The following enables that url parameters can be written with or without / at the end
            for param, value in self.CONFIG_PARAMS.items():
                if isinstance(value, str) and value.startswith('http'):
                    config[param] = config[param].rstrip('/')

            return config

        @staticmethod
        def get_config_file():
            """ Checks if configuration file exists and returns its file path

            :return: location of configuration file
            :rtype: str
            """
            config_file = os.path.join(os.path.dirname(__file__), 'config.json')

            if not os.path.isfile(config_file):
                raise IOError('Configuration file does not exist: %s' % os.path.abspath(config_file))

            return config_file

        def load_configuration(self):
            """ Method reads and loads the configuration file.
            """
            with open(self.get_config_file(), 'r') as cfg_file:
                config = json.load(cfg_file)

            config = self._parse_configuration(config)

            for prop in config:
                if prop in self.CONFIG_PARAMS:
                    setattr(self, prop, config[prop])

        def get_config(self):
            """ Returns ordered dictionary with configuration parameters

            :return: Ordered dictionary
            :rtype: collections.OrderedDict
            """
            config = OrderedDict((prop, getattr(self, prop)) for prop in self.CONFIG_PARAMS)
            if config['instance_id'] is None:
                config['instance_id'] = ''
            return config

        def save_configuration(self):
            """ Method saves changed parameter values to the configuration file.
            """
            config = self.get_config()

            self._parse_configuration(config)

            with open(self.get_config_file(), 'w') as cfg_file:
                json.dump(config, cfg_file, indent=2)

    _instance = None

    def __init__(self):
        if not SHConfig._instance:
            SHConfig._instance = self._SHConfig()

        for prop in self._instance.CONFIG_PARAMS:
            setattr(self, prop, getattr(self._instance, prop))

    def __getattr__(self, name):
        """ This is called only if the class doesn't have the attribute itself
        """
        return getattr(self._instance, name)

    def __getitem__(self, name):
        return getattr(self, name)

    def __dir__(self):
        return sorted(list(dir(super())) + list(self._instance.CONFIG_PARAMS))

    def __str__(self):
        """ Content of SHConfig in json schema
        """
        return json.dumps(self.get_config_dict(), indent=2)

    def __repr__(self):
        """ Representation of SHConfig parameters
        """
        repr_list = ['{}('.format(self.__class__.__name__)]

        for key, value in self.get_config_dict().items():
            repr_list.append('%s=%r,' % (key, value))

        return '\n  '.join(repr_list).strip(',') + '\n)'

    def save(self):
        """ Method that saves configuration parameter changes from instance of SHConfig class to global config class and
        to `config.json` file.

        :Example:
            ``my_config = SHConfig()`` \n
            ``my_config.instance_id = '<new instance id>'`` \n
            ``my_config.save()``
        """
        is_changed = False
        for prop in self._instance.CONFIG_PARAMS:
            if getattr(self, prop) != getattr(self._instance, prop):
                is_changed = True
                setattr(self._instance, prop, getattr(self, prop))
        if is_changed:
            self._instance.save_configuration()

    def reset(self, params=...):
        """ Resets configuration class to initial values. Use `SHConfig.save()` method in order to save this change.

        :param params: Parameters which will be reset. Parameters can be specified with a list of names, e.g.
            ``['instance_id', 'aws_access_key_id', 'aws_secret_access_key']``, or as a single name, e.g.
            ``'sh_base_url'``. By default all parameters will be reset and default value is ``Ellipsis``.
        :type params: Ellipsis or list(str) or str
        """
        if params is ...:
            params = self.get_params()
        if isinstance(params, str):
            self._reset_param(params)
        elif isinstance(params, (list, tuple)):
            for param in params:
                self._reset_param(param)
        else:
            raise ValueError('Parameters must be specified in form of a list of strings or as a single string, instead '
                             'got {}'.format(params))

    def _reset_param(self, param):
        """ Resets a single parameter

        :param param: A configuration parameter
        :type param: str
        """
        if param not in self._instance.CONFIG_PARAMS:
            raise ValueError("Cannot reset unknown parameter '{}'".format(param))
        setattr(self, param, self._instance.CONFIG_PARAMS[param])

    def get_params(self):
        """ Returns a list of parameter names

        :return: List of parameter names
        :rtype: list(str)
        """
        return list(self._instance.CONFIG_PARAMS)

    def get_config_dict(self):
        """ Get a dictionary representation of `SHConfig` class

        :return: A dictionary with configuration parameters
        :rtype: OrderedDict
        """
        return OrderedDict((prop, getattr(self, prop)) for prop in self._instance.CONFIG_PARAMS)

    def get_config_location(self):
        """ Returns location of configuration file on disk

        :return: File path of `config.json` file
        :rtype: str
        """
        return self._instance.get_config_file()

    def has_eocloud_url(self):
        """ Checks if base Sentinel Hub URL is set to eocloud URL

        :return: `True` if 'eocloud' string is in base OGC URL else `False`
        :rtype: bool
        """
        return 'eocloud' in self.sh_base_url

    def get_sh_oauth_url(self):
        """ Provides URL for Sentinel Hub authentication endpoint

        :return: An URL endpoint
        :rtype: str
        """
        return '{}/oauth/token'.format(self.sh_base_url)

    def get_sh_processing_api_url(self):
        """  Provides URL for Sentinel Hub processing API endpoint

        :return: An URL endpoint
        :rtype: str
        """
        return '{}/api/v1/process'.format(self.sh_base_url)

    def get_sh_ogc_url(self):
        """ Provides URL for Sentinel Hub OGC endpoint

        :return: An URL endpoint
        :rtype: str
        """
        ogc_enpoint = 'v1' if self.has_eocloud_url() else 'ogc'
        return '{}/{}'.format(self.sh_base_url, ogc_enpoint)

    def get_sh_rate_limit_url(self):
        """ Provides URL for Sentinel Hub rate limiting endpoint

        :return: An URL endpoint
        :rtype: str
        """
        return '{}/aux/ratelimit'.format(self.sh_base_url)



"""
Module implementing geometry classes
"""
import warnings
from abc import ABC, abstractmethod
from math import ceil

import shapely.ops
import shapely.geometry
import shapely.wkt

from .constants import CRS
from .exceptions import SHDeprecationWarning
from .geo_utils import transform_point


class BaseGeometry(ABC):
    """ Base geometry class
    """
    def __init__(self, crs):
        """
        :param crs: Coordinate reference system of the geometry
        :type crs: constants.CRS
        """
        self._crs = CRS(crs)

    def _repr_svg_(self):
        """ Using shapely's svg geometry visualization for Jupyter notebooks
        """
        # pylint: disable=protected-access
        return self.geometry._repr_svg_()

    @property
    def crs(self):
        """ Returns the coordinate reference system (CRS)

        :return: Coordinate reference system Enum
        :rtype: constants.CRS
        """
        return self._crs

    def get_crs(self):
        """ Returns the coordinate reference system (CRS)

        :return: Coordinate reference system Enum
        :rtype: constants.CRS
        """
        message = 'This method will be removed in version 3.0, use property {}.crs ' \
                  'instead'.format(self.__class__.__name__)
        warnings.warn(message, category=SHDeprecationWarning)
        return self.crs

    @property
    @abstractmethod
    def geometry(self):
        """ An abstract property - ever subclass must implement geometry property
        """
        raise NotImplementedError

    def get_geometry(self):
        """ Returns shapely geometry

        :return: A polygon or multipolygon
        :rtype: shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        """
        message = 'This method will be removed in version 3.0, use property {}.geometry ' \
                  'instead'.format(self.__class__.__name__)
        warnings.warn(message, category=SHDeprecationWarning)
        return self.geometry

    @property
    def geojson(self):
        """ Returns representation in a GeoJSON format. Use json.dump for writing it to file.

        :return: A dictionary in GeoJSON format
        :rtype: dict
        """
        return {
            **self._crs_to_geojson(),
            **shapely.geometry.mapping(self.geometry)
        }

    def get_geojson(self):
        """ Returns representation in a GeoJSON format. Use json.dump for writing it to file.

        :return: A dictionary in GeoJSON format
        :rtype: dict
        """
        return self.geojson

    def _crs_to_geojson(self):
        """ Helper method which generates part of GeoJSON format related to CRS
        """
        return {
            'crs': {
                'type': 'name',
                'properties': {'name': 'urn:ogc:def:crs:EPSG::{}'.format(self.crs.value)}
            }
        }

    @property
    def wkt(self):
        """ Transforms geometry object into `Well-known text` format

        :return: string in WKT format
        :rtype: str
        """
        return self.geometry.wkt


class BBox(BaseGeometry):
    """ Class representing a bounding box in a given CRS.

    Throughout the sentinelhub package this class serves as the canonical representation of a bounding
    box. It can initialize itself from multiple representations:

        1) ``((min_x,min_y),(max_x,max_y))``,
        2) ``(min_x,min_y,max_x,max_y)``,
        3) ``[min_x,min_y,max_x,max_y]``,
        4) ``[[min_x, min_y],[max_x,max_y]]``,
        5) ``[(min_x, min_y),(max_x,max_y)]``,
        6) ``([min_x, min_y],[max_x,max_y])``,
        7) ``'min_x,min_y,max_x,max_y'``,
        8) ``{'min_x':min_x, 'max_x':max_x, 'min_y':min_y, 'max_y':max_y}``,
        9) ``bbox``, where ``bbox`` is an instance of ``BBox``.

    Note that BBox coordinate system depends on ``crs`` parameter:
        - In case of ``constants.CRS.WGS84`` axis x represents longitude and axis y represents latitude
        - In case of ``constants.CRS.POP_WEB`` axis x represents easting and axis y represents northing
        - In case of ``constants.CRS.UTM_*`` axis x represents easting and axis y represents northing
    """
    def __init__(self, bbox, crs):
        """
        :param bbox: A bbox in any valid representation
        :param crs: Coordinate reference system of the bounding box
        :type crs: constants.CRS
        """
        x_fst, y_fst, x_snd, y_snd = BBox._to_tuple(bbox)
        self.min_x = min(x_fst, x_snd)
        self.max_x = max(x_fst, x_snd)
        self.min_y = min(y_fst, y_snd)
        self.max_y = max(y_fst, y_snd)

        super().__init__(crs)

    def __iter__(self):
        """ This method enables iteration over coordinates of bounding box
        """
        return iter(self.lower_left + self.upper_right)

    def __repr__(self):
        """ Class representation
        """
        return '{}((({}, {}), ({}, {})), crs={})'.format(self.__class__.__name__, *self, self.crs)

    def __str__(self, reverse=False):
        """ Transforms bounding box into a string of coordinates

        :param reverse: `True` if x and y coordinates should be switched and `False` otherwise
        :type reverse: bool
        :return: String of coordinates
        :rtype: str
        """
        if reverse:
            return '{},{},{},{}'.format(self.min_y, self.min_x, self.max_y, self.max_x)
        return '{},{},{},{}'.format(self.min_x, self.min_y, self.max_x, self.max_y)

    def __eq__(self, other):
        """ Method for comparing two bounding boxes

        :param other: Another bounding box object
        :type other: BBox
        :return: `True` if bounding boxes have the same coordinates and the same CRS and `False otherwise
        :rtype: bool
        """
        return list(self) == list(other) and self.crs == other.crs

    @property
    def lower_left(self):
        """ Returns the lower left vertex of the bounding box

        :return: min_x, min_y
        :rtype: (float, float)
        """
        return self.min_x, self.min_y

    def get_lower_left(self):
        """ Returns the lower left vertex of the bounding box

        :return: min_x, min_y
        :rtype: (float, float)
        """
        message = 'This method will be removed in version 3.0, use property BBox.lower_left instead'
        warnings.warn(message, category=SHDeprecationWarning)
        return self.lower_left

    @property
    def upper_right(self):
        """ Returns the upper right vertex of the bounding box

        :return: max_x, max_y
        :rtype: (float, float)
        """
        return self.max_x, self.max_y

    def get_upper_right(self):
        """ Returns the upper right vertex of the bounding box

        :return: max_x, max_y
        :rtype: (float, float)
        """
        message = 'This method will be removed in version 3.0, use property BBox.upper_right instead'
        warnings.warn(message, category=SHDeprecationWarning)
        return self.upper_right

    @property
    def middle(self):
        """ Returns the middle point of the bounding box

        :return: middle point
        :rtype: (float, float)
        """
        return (self.min_x + self.max_x) / 2, (self.min_y + self.max_y) / 2

    def get_middle(self):
        """ Returns the middle point of the bounding box

        :return: middle point
        :rtype: (float, float)
        """
        message = 'This method will be removed in version 3.0, use property BBox.middle instead'
        warnings.warn(message, category=SHDeprecationWarning)
        return self.middle

    def reverse(self):
        """ Returns a new BBox object where x and y coordinates are switched

        :return: New BBox object with switched coordinates
        :rtype: BBox
        """
        return BBox((self.min_y, self.min_x, self.max_y, self.max_x), crs=self.crs)

    def transform(self, crs):
        """ Transforms BBox from current CRS to target CRS

        :param crs: target CRS
        :type crs: constants.CRS
        :return: Bounding box in target CRS
        :rtype: BBox
        """
        new_crs = CRS(crs)
        return BBox((transform_point(self.lower_left, self.crs, new_crs),
                     transform_point(self.upper_right, self.crs, new_crs)), crs=new_crs)

    def buffer(self, buffer):
        """ Changes both BBox dimensions (width and height) by a percentage of size of each dimension. If number is
        negative, the size will decrease. Returns a new instance of BBox object.

        :param buffer: A percentage of BBox size change
        :type buffer: float
        :return: A new bounding box of buffered size
        :rtype: BBox
        """
        if buffer < -1:
            raise ValueError('Cannot reduce the bounding box to nothing, buffer must be >= -1.0')
        ratio = 1 + buffer
        mid_x, mid_y = self.middle
        return BBox((mid_x - (mid_x - self.min_x) * ratio, mid_y - (mid_y - self.min_y) * ratio,
                     mid_x + (self.max_x - mid_x) * ratio, mid_y + (self.max_y - mid_y) * ratio), self.crs)

    def get_polygon(self, reverse=False):
        """ Returns a tuple of coordinates of 5 points describing a polygon. Points are listed in clockwise order, first
        point is the same as the last.

        :param reverse: `True` if x and y coordinates should be switched and `False` otherwise
        :type reverse: bool
        :return: `((x_1, y_1), ... , (x_5, y_5))`
        :rtype: tuple(tuple(float))
        """
        bbox = self.reverse() if reverse else self
        polygon = ((bbox.min_x, bbox.min_y),
                   (bbox.min_x, bbox.max_y),
                   (bbox.max_x, bbox.max_y),
                   (bbox.max_x, bbox.min_y),
                   (bbox.min_x, bbox.min_y))
        return polygon

    @property
    def geometry(self):
        """ Returns polygon geometry in shapely format

        :return: A polygon in shapely format
        :rtype: shapely.geometry.polygon.Polygon
        """
        return shapely.geometry.Polygon(self.get_polygon())

    def get_partition(self, num_x=None, num_y=None, size_x=None, size_y=None):
        """ Partitions bounding box into smaller bounding boxes of the same size.

        If `num_x` and `num_y` are specified, the total number of BBoxes is know but not the size. If `size_x` and
        `size_y` are provided, the BBox size is fixed but the number of BBoxes is not known in advance. In the latter
        case, the generated bounding boxes might cover an area larger than the parent BBox.

        :param num_x: Number of parts BBox will be horizontally divided into.
        :type num_x: int or None
        :param num_y: Number of parts BBox will be vertically divided into.
        :type num_y: int or None
        :param size_x: Physical dimension of BBox along easting coordinate
        :type size_x: float or None
        :param size_y: Physical dimension of BBox along northing coordinate
        :type size_y: float or None
        :return: Two-dimensional list of smaller bounding boxes. Their location is
        :rtype: list(list(BBox))
        """
        if (num_x is not None and num_y is not None) and (size_x is None and size_y is None):
            size_x, size_y = (self.max_x - self.min_x) / num_x, (self.max_y - self.min_y) / num_y
        elif (size_x is not None and size_y is not None) and (num_x is None and num_y is None):
            num_x, num_y = ceil((self.max_x - self.min_x) / size_x), ceil((self.max_y - self.min_y) / size_y)
        else:
            raise ValueError('Not supported partition. Either (num_x, num_y) or (size_x, size_y) must be specified')

        return [[BBox([self.min_x + i * size_x, self.min_y + j * size_y,
                       self.min_x + (i + 1) * size_x, self.min_y + (j + 1) * size_y],
                      crs=self.crs) for j in range(num_y)] for i in range(num_x)]

    def get_transform_vector(self, resx, resy):
        """ Given resolution it returns a transformation vector

        :param resx: Resolution in x direction
        :type resx: float or int
        :param resy: Resolution in y direction
        :type resy: float or int
        :return: A tuple with 6 numbers representing transformation vector
        :rtype: tuple(float)
        """
        return self.min_x, self._parse_resolution(resx), 0, self.max_y, 0, -self._parse_resolution(resy)

    @staticmethod
    def _parse_resolution(res):
        """ Helper method for parsing given resolution. It will also try to parse a string into float

        :return: A float value of resolution
        :rtype: float
        """
        if isinstance(res, str):
            return float(res.strip('m'))
        if isinstance(res, (int, float)):
            return float(res)

        raise TypeError('Resolution should be a float, got resolution of type {}'.format(type(res)))

    @staticmethod
    def _to_tuple(bbox):
        """ Converts the input bbox representation (see the constructor docstring for a list of valid representations)
        into a flat tuple

        :param bbox: A bbox in one of 7 forms listed in the class description.
        :return: A flat tuple of size
        :raises: TypeError
        """
        if isinstance(bbox, (list, tuple)):
            return BBox._tuple_from_list_or_tuple(bbox)
        if isinstance(bbox, str):
            return BBox._tuple_from_str(bbox)
        if isinstance(bbox, dict):
            return BBox._tuple_from_dict(bbox)
        if isinstance(bbox, BBox):
            return BBox._tuple_from_bbox(bbox)
        if isinstance(bbox, shapely.geometry.base.BaseGeometry):
            return bbox.bounds
        raise TypeError('Invalid bbox representation')

    @staticmethod
    def _tuple_from_list_or_tuple(bbox):
        """ Converts a list or tuple representation of a bbox into a flat tuple representation.

        :param bbox: a list or tuple with 4 coordinates that is either flat or nested
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: TypeError
        """
        if len(bbox) == 4:
            return tuple(map(float, bbox))
        if len(bbox) == 2 and all([isinstance(point, (list, tuple)) for point in bbox]):
            return BBox._tuple_from_list_or_tuple(bbox[0] + bbox[1])
        raise TypeError('Expected a valid list or tuple representation of a bbox')

    @staticmethod
    def _tuple_from_str(bbox):
        """ Parses a string of numbers separated by any combination of commas and spaces

        :param bbox: e.g. str of the form `min_x ,min_y  max_x, max_y`
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return tuple([float(s) for s in bbox.replace(',', ' ').split() if s])

    @staticmethod
    def _tuple_from_dict(bbox):
        """ Converts a dictionary representation of a bbox into a flat tuple representation

        :param bbox: a dict with keys "min_x, "min_y", "max_x", and "max_y"
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: KeyError
        """
        return bbox['min_x'], bbox['min_y'], bbox['max_x'], bbox['max_y']

    @staticmethod
    def _tuple_from_bbox(bbox):
        """ Converts a BBox instance into a tuple

        :param bbox: An instance of the BBox type
        :return: tuple (min_x, min_y, max_x, max_y)
        """
        return bbox.lower_left + bbox.upper_right


class Geometry(BaseGeometry):
    """ A class that combines shapely geometry with coordinate reference system. It currently supports polygons and
    multipolygons.

    It can be initialize with any of the following geometry representations:
    - `shapely.geometry.Polygon` or `shapely.geometry.MultiPolygon`
    - A GeoJSON dictionary with (multi)polygon coordinates
    - A WKT string with (multi)polygon coordinates
    """
    def __init__(self, geometry, crs):
        """
        :param geometry: A polygon or multipolygon in any valid representation
        :type geometry: shapely.geometry.Polygon or shapely.geometry.MultiPolygon or dict or str
        :param crs: Coordinate reference system of the geometry
        :type crs: constants.CRS
        """
        self._geometry = self._parse_geometry(geometry)

        super().__init__(crs)

    def __repr__(self):
        """ Method for class representation
        """
        return '{}({}, crs={})'.format(self.__class__.__name__, self.wkt, self.crs)

    def __eq__(self, other):
        """ Method for comparing two Geometry classes

        :param other: Another Geometry object
        :type other: Geometry
        :return: `True` if geometry objects have the same geometry and CRS and `False` otherwise
        :rtype: bool
        """
        return self.geometry == other.geometry and self.crs == other.crs

    def reverse(self):
        """ Returns a new Geometry object where x and y coordinates are switched

        :return: New Geometry object with switched coordinates
        :rtype: Geometry
        """
        return Geometry(shapely.ops.transform(lambda x, y: (y, x), self.geometry), crs=self.crs)

    def transform(self, crs):
        """ Transforms Geometry from current CRS to target CRS

        :param crs: target CRS
        :type crs: constants.CRS
        :return: Geometry in target CRS
        :rtype: Geometry
        """
        new_crs = CRS(crs)

        geometry = self.geometry
        if new_crs is not self.crs:
            transform_function = self.crs.get_transform_function(new_crs)
            geometry = shapely.ops.transform(transform_function, geometry)

        return Geometry(geometry, crs=new_crs)

    @property
    def geometry(self):
        """ Returns shapely object representing geometry in this class

        :return: A polygon or a multipolygon in shapely format
        :rtype: shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        """
        return self._geometry

    @property
    def bbox(self):
        """ Returns BBox object representing bounding box around the geometry

        :return: A bounding box, with same CRS
        :rtype: BBox
        """
        return BBox(self.geometry, self.crs)

    @staticmethod
    def _parse_geometry(geometry):
        """ Parses given geometry into shapely object

        :param geometry:
        :return: Shapely polygon or multipolygon
        :rtype: shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        :raises TypeError
        """
        if isinstance(geometry, str):
            geometry = shapely.wkt.loads(geometry)
        elif isinstance(geometry, dict):
            geometry = shapely.geometry.shape(geometry)
        elif not isinstance(geometry, shapely.geometry.base.BaseGeometry):
            raise TypeError('Unsupported geometry representation')

        if not isinstance(geometry, (shapely.geometry.Polygon, shapely.geometry.MultiPolygon)):
            raise ValueError('Supported geometry types are polygon and multipolygon, got {}'.format(type(geometry)))

        return geometry


class BBoxCollection(BaseGeometry):
    """ A collection of bounding boxes
    """
    def __init__(self, bbox_list):
        """
        :param bbox_list: A list of BBox objects which have to be in the same CRS
        :type bbox_list: list(BBox)
        """
        self._bbox_list, crs = self._parse_bbox_list(bbox_list)
        self._geometry = self._get_geometry()

        super().__init__(crs)

    def __repr__(self):
        """ Method for class representation
        """
        return '{}({})'.format(self.__class__.__name__, ', '.join([repr(bbox) for bbox in self.bbox_list]))

    def __eq__(self, other):
        """ Method for comparing two BBoxCollection classes
        """
        return self.crs == other.crs and len(self.bbox_list) == len(other.bbox_list) and \
            all(bbox == other_bbox for bbox, other_bbox in zip(self, other))

    def __iter__(self):
        """ This method enables iteration over bounding boxes in collection
        """
        return iter(self.bbox_list)

    @property
    def bbox_list(self):
        """ Returns the list of bounding boxes from collection

        :return: The list of bounding boxes
        :rtype: list(BBox)
        """
        return self._bbox_list

    @property
    def geometry(self):
        """ Returns shapely object representing geometry

        :return: A multipolygon of bounding boxes
        :rtype: shapely.geometry.MultiPolygon
        """
        return self._geometry

    @property
    def bbox(self):
        """ Returns BBox object representing bounding box around the geometry

        :return: A bounding box, with same CRS
        :rtype: BBox
        """
        return BBox(self.geometry, self.crs)

    def reverse(self):
        """ Returns a new BBoxCollection object where all x and y coordinates are switched

        :return: New Geometry object with switched coordinates
        :rtype: BBoxCollection
        """
        return BBoxCollection([bbox.reverse() for bbox in self.bbox_list])

    def transform(self, crs):
        """ Transforms BBoxCollection from current CRS to target CRS

        :param crs: target CRS
        :type crs: constants.CRS
        :return: BBoxCollection in target CRS
        :rtype: BBoxCollection
        """
        return BBoxCollection([bbox.transform(crs) for bbox in self.bbox_list])

    def _get_geometry(self):
        """ Creates a multipolygon of bounding box polygons
        """
        return shapely.geometry.MultiPolygon([bbox.geometry for bbox in self.bbox_list])

    @staticmethod
    def _parse_bbox_list(bbox_list):
        """ Helper method for parsing a list of bounding boxes
        """
        if isinstance(bbox_list, BBoxCollection):
            return bbox_list.bbox_list, bbox_list.crs

        if not isinstance(bbox_list, list) or not bbox_list:
            raise ValueError('Expected non-empty list of BBox objects')

        for bbox in bbox_list:
            if not isinstance(bbox, BBox):
                raise ValueError('Elements in the list should be of type {}, got {}'.format(BBox.__name__, type(bbox)))

        crs = bbox_list[0].crs
        for bbox in bbox_list:
            if bbox.crs is not crs:
                raise ValueError('All bounding boxes should have the same CRS')

        return bbox_list, crs



"""
Module for working with Sentinel Hub FIS service
"""
import logging

from .time_utils import parse_time_interval
from .download import DownloadRequest
from .constants import MimeType, CRS, SHConstants, RequestType
from .ogc import OgcImageService
from .geometry import Geometry

LOGGER = logging.getLogger(__name__)


class FisService(OgcImageService):
    """ Sentinel Hub OGC services class for providing FIS data

    Intermediate layer between FIS requests and the Sentinel Hub FIS services.
    """
    def get_request(self, request):
        """ Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: OGC-type request with specified bounding box, time interval, and cloud coverage for specific
            product.
        :type request: OgcRequest or GeopediaRequest
        :return: list of DownloadRequests
        """

        return [self._create_request(request=request, geometry=geometry) for geometry in request.geometry_list]

    def _create_request(self, request, geometry):

        url = self.get_base_url(request)
        authority = self.instance_id if hasattr(self, 'instance_id') else request.theme
        headers = {'Content-Type': MimeType.JSON.get_string(), **SHConstants.HEADERS}

        post_data = {**self._get_common_url_parameters(request), **self._get_fis_parameters(request, geometry)}
        post_data = {k.lower(): v for k, v in post_data.items()}  # lowercase required on SH service

        return DownloadRequest(url='{}/{}'.format(url, authority),
                               post_values=post_data,
                               filename=self.get_filename(request, geometry),
                               data_type=MimeType.JSON, headers=headers,
                               request_type=RequestType.POST)

    @staticmethod
    def get_filename(request, geometry):
        """ Returns the filename location on disk where data is or is going to be stored.
        The files are stored in the folder specified by the user when initialising OGC-type
        of request. The name of the file has the following structure:

        {service_type}_{layer}_{geometry}_{crs}_{start_time}_{end_time}_{resolution}_{bins}_{histogram_type}_
        \\*{custom_url_params}.json

        :param request: FIS request
        :type request: FisRequest
        :param geometry: geometry object
        :type geometry: BBox or Geometry
        :return: filename for this request
        :rtype: str
        """
        date_interval = parse_time_interval(request.time)
        geometry_string = geometry.wkt if isinstance(geometry, Geometry) else str(geometry)

        filename = '_'.join([
            str(request.service_type.value),
            request.layer,
            geometry_string,
            CRS.ogc_string(geometry.crs),
            '{}_{}'.format(date_interval[0], date_interval[1]),
            request.resolution,
            str(request.bins) if request.bins else '',
            request.histogram_type.value if request.histogram_type else ''
        ])

        filename = OgcImageService.filename_add_custom_url_params(filename, request)

        return OgcImageService.finalize_filename(filename, MimeType.JSON)



"""
Module implementing error handlers which can occur during download procedure
"""
import logging
import time

import requests

from ..decoding import decode_sentinelhub_err_msg
from ..exceptions import DownloadFailedException


LOGGER = logging.getLogger(__name__)


def fail_user_errors(download_func):
    """ Decorator function for handling user errors
    """

    def new_download_func(self, request):
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if exception.response.status_code < requests.status_codes.codes.INTERNAL_SERVER_ERROR and \
                exception.response.status_code != requests.status_codes.codes.TOO_MANY_REQUESTS:

                raise DownloadFailedException(_create_download_failed_message(exception, request.url)) from exception
            raise exception from exception

    return new_download_func


def retry_temporal_errors(download_func):
    """ Decorator function for handling server and connection errors
    """
    backoff_coefficient = 3

    def new_download_func(self, request):
        download_attempts = self.config.max_download_attempts
        sleep_time = self.config.download_sleep_time

        for attempt_num in range(download_attempts):
            try:
                return download_func(self, request)
            except requests.RequestException as exception:

                if not (_is_temporal_problem(exception) or
                        (isinstance(exception, requests.HTTPError) and
                         exception.response.status_code >= requests.status_codes.codes.INTERNAL_SERVER_ERROR)):
                    raise exception from exception

                if attempt_num == download_attempts - 1:
                    raise DownloadFailedException(_create_download_failed_message(exception, request.url)) \
                        from exception

                LOGGER.debug('Download attempt failed: %s\n%d attempts left, will retry in %ds', exception,
                             download_attempts - attempt_num - 1, sleep_time)
                time.sleep(sleep_time)
                sleep_time *= backoff_coefficient

    return new_download_func


def fail_missing_file(download_func):
    """ A decorator for raising an error if a file is missing
    """
    def new_download_func(self, request):
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if exception.response.status_code == requests.status_codes.codes.NOT_FOUND:
                raise DownloadFailedException('File in location %s is missing' % request.url) from exception

            raise exception from exception

    return new_download_func


def _is_temporal_problem(exception):
    """ Checks if the obtained exception is temporal and if download attempt should be repeated

    :param exception: Exception raised during download
    :type exception: Exception
    :return: `True` if exception is temporal and `False` otherwise
    :rtype: bool
    """
    try:
        return isinstance(exception, (requests.ConnectionError, requests.Timeout))
    except AttributeError:  # Earlier requests versions might not have requests.Timeout
        return isinstance(exception, requests.ConnectionError)


def _create_download_failed_message(exception, url):
    """ Creates message describing why download has failed

    :param exception: Exception raised during download
    :type exception: Exception
    :param url: An URL from where download was attempted
    :type url: str
    :return: Error message
    :rtype: str
    """
    message = 'Failed to download from:\n{}\nwith {}:\n{}'.format(url, exception.__class__.__name__, exception)

    if _is_temporal_problem(exception):
        if isinstance(exception, requests.ConnectionError):
            message += '\nPlease check your internet connection and try again.'
        else:
            message += '\nThere might be a problem in connection or the server failed to process ' \
                       'your request. Please try again.'
    elif isinstance(exception, requests.HTTPError):
        server_message = decode_sentinelhub_err_msg(exception.response)
        message += '\nServer response: "{}"'.format(server_message)

    return message



"""
Module implementing download client that is adjusted to download from AWS
"""
import logging

import boto3
from botocore.exceptions import NoCredentialsError

from ..exceptions import AwsDownloadFailedException
from .client import DownloadClient, get_json, get_xml
from .handlers import fail_missing_file


LOGGER = logging.getLogger(__name__)


class AwsDownloadClient(DownloadClient):
    """ An AWS download client class
    """

    GLOBAL_S3_CLIENT = None

    @fail_missing_file
    def _execute_download(self, request):

        if not self.is_s3_request(request):
            return super()._execute_download(request)

        s3_client = self._get_s3_client()

        response = self._do_download(request, s3_client)
        response_content = response['Body'].read()

        LOGGER.debug('Successful download from %s', request.url)
        return response_content

    def _get_s3_client(self):

        key_args = {}
        if self.config.aws_access_key_id and self.config.aws_secret_access_key:
            key_args = {
                'aws_access_key_id': self.config.aws_access_key_id,
                'aws_secret_access_key': self.config.aws_secret_access_key
            }

        try:
            # check if it would be better to just take the global one
            s3_client = boto3.Session().client('s3', **key_args)
            AwsDownloadClient.GLOBAL_S3_CLIENT = s3_client  # Storing the client prevents warning about unclosed socket
        except KeyError:  # Sometimes creation of client fails and we use the global client if it exists
            if AwsDownloadClient.GLOBAL_S3_CLIENT is None:
                raise ValueError('Failed to create a client for download from AWS')
            s3_client = AwsDownloadClient.GLOBAL_S3_CLIENT

        return s3_client

    @staticmethod
    def _do_download(request, s3_client):
        _, _, bucket_name, url_key = request.url.split('/', 3)

        # include:
        # https://github.com/sentinel-hub/sentinelhub-py/blob/feat/aws-speedup/sentinelhub/download.py#L310

        try:
            return s3_client.get_object(Bucket=bucket_name, Key=url_key, RequestPayer='requester')
        except NoCredentialsError:
            raise ValueError(
                'The requested data is in Requester Pays AWS bucket. In order to download the data please set '
                'your access key either in AWS credentials file or in sentinelhub config.json file using '
                'command line:\n'
                '$ sentinelhub.config --aws_access_key_id <your AWS key> --aws_secret_access_key '
                '<your AWS secret key>')
        except s3_client.exceptions.NoSuchKey:
            raise AwsDownloadFailedException('File in location %s is missing' % request.url)
        except s3_client.exceptions.NoSuchBucket:
            raise ValueError('Aws bucket %s does not exist' % bucket_name)

    @staticmethod
    def is_s3_request(request):
        """ Checks if data has to be downloaded from AWS s3 bucket

        :return: `True` if url describes location at AWS s3 bucket and `False` otherwise
        :rtype: bool
        """
        return request.url.startswith('s3://')


def get_aws_json(*args, **kwargs):
    """ Download a json from AWS
    """
    return get_json(*args, download_client_class=AwsDownloadClient, **kwargs)


def get_aws_xml(*args, **kwargs):
    """ Download an xml from AWS
    """
    return get_xml(*args, download_client_class=AwsDownloadClient, **kwargs)



"""
Module defining custom package exceptions
"""

import warnings


class BaseSentinelHubException(Exception):
    """ A base package exception
    """


class DownloadFailedException(BaseSentinelHubException):
    """ General exception which is raised whenever download fails
    """


class OutOfRequestsException(DownloadFailedException):
    """ An exception raised whenever download cannot be done because user has run out of requests or processing units
    """


class AwsDownloadFailedException(DownloadFailedException):
    """ This exception is raised when download fails because of a missing file in AWS
    """


class ImageDecodingError(BaseSentinelHubException):
    """ This exception is raised when downloaded image is not properly decoded
    """


class SHDeprecationWarning(DeprecationWarning):
    """ A custom deprecation warning for sentinelhub-py package
    """


class SHUserWarning(UserWarning):
    """ A custom user warning for sentinelhub-py package
    """


class SHRuntimeWarning(RuntimeWarning):
    """ A custom runtime warning for sentinelhub-py package, it should be
    """


warnings.simplefilter('default', SHDeprecationWarning)
warnings.simplefilter('default', SHUserWarning)
warnings.simplefilter('always', SHRuntimeWarning)



"""
The required empty module
"""

from .request import DownloadRequest
from .client import DownloadClient, get_json, get_xml
from .sentinelhub_client import SentinelHubDownloadClient
from .aws_client import AwsDownloadClient



"""
Module implementing a rate-limited multi-threaded download client for downloading from Sentinel Hub service
"""
import logging
import sys
import warnings
import time
from threading import Lock

import requests

from .handlers import fail_user_errors, retry_temporal_errors
from .client import DownloadClient
from ..exceptions import SHUserWarning
from ..sentinelhub_session import SentinelHubSession
from ..sentinelhub_rate_limit import SentinelHubRateLimit


LOGGER = logging.getLogger(__name__)


class SentinelHubDownloadClient(DownloadClient):
    """ Download client specifically configured for download from Sentinel Hub service
    """
    def __init__(self, *, session=None, **kwargs):
        """
        :param session: An OAuth2 session with Sentinel Hub service
        :type session: SentinelHubSession or None
        :param kwargs: Optional parameters from DownloadClient
        """
        super().__init__(**kwargs)

        self.session = self._configure_session(session)

        self.rate_limit = SentinelHubRateLimit(self.session)
        self.lock = Lock()

    def _configure_session(self, session):
        """ Configures session object if credentials are given
        """
        if isinstance(session, SentinelHubSession):
            return session

        if session is None:
            if self.config.sh_client_id and self.config.sh_client_secret:
                return SentinelHubSession(config=self.config)

            message = "In order to achieve faster download performance please set configuration parameters " \
                      "'sh_client_id' and 'sh_client_secret'"
            warnings.warn(message, category=SHUserWarning)

            return None

        raise ValueError('Given session should be an instance of SentinelHubSession')

    @retry_temporal_errors
    @fail_user_errors
    def _execute_download(self, request):

        while True:
            sleep_time = self._execute_with_lock(self.rate_limit.register_next)

            if sleep_time == 0:
                response = self._do_download(request)

                self._execute_with_lock(self.rate_limit.update, response.headers)

                if response.status_code != requests.status_codes.codes.TOO_MANY_REQUESTS:
                    response.raise_for_status()

                    LOGGER.debug('Successful download from %s', request.url)
                    return response.content
            else:
                time.sleep(sleep_time)

    def _execute_with_lock(self, thread_unsafe_function, *args, **kwargs):
        """ Executes a function inside a thread lock and handles potential errors
        """
        self.lock.acquire()
        exception = None
        try:
            result = thread_unsafe_function(*args, **kwargs)
        except BaseException as exc:
            traceback = sys.exc_info()[2]
            exception = exc
        self.lock.release()

        if exception is not None:
            raise exception.with_traceback(traceback)

        return result

    def _do_download(self, request):
        """ Runs the download
        """
        return requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=self._prepare_headers(request.headers),
            timeout=self.config.download_timeout_seconds
        )

    def _prepare_headers(self, headers):
        """ Prepares final headers by potentially joining them with session headers
        """
        if self.session is None:
            return headers

        return {
            **self.session.session_headers,
            **headers
        }



"""
Module implementing the main download client class
"""
import concurrent.futures
import logging
import warnings
import os
import sys

import requests

from ..config import SHConfig
from ..constants import RequestType, MimeType
from ..decoding import decode_data
from ..exceptions import DownloadFailedException, SHRuntimeWarning
from ..io_utils import LocalIO
from .handlers import fail_user_errors, retry_temporal_errors
from .request import DownloadRequest
from .cache import hash_request


LOGGER = logging.getLogger(__name__)


class DownloadClient:
    """ The download client class which takes care of downloading the data, parallelizing the download through threads
    and catching any exceptions.
    """
    def __init__(self, *, redownload=False, raise_download_errors=True, io_object=None, config=None):
        """
        :param redownload: If `True` the data will always be downloaded again. By default this is set to `False` and
            the data that has already been downloaded and saved to an expected location will be read from the
            location instead of being downloaded again.
        :type redownload: bool
        :param raise_download_errors: If `True` any error in download process will be raised as
            `DownloadFailedException`. If `False` failed downloads will only raise warnings.
        :type raise_download_errors: bool
        :param io_object: An instance of IO class
        :type io_object: LocalIO
        :param config: An instance of configuration class
        :type config: SHConfig
        """
        self.redownload = redownload
        self.raise_download_errors = raise_download_errors

        self.io_object = LocalIO() if io_object is None else io_object

        self.config = SHConfig() if config is None else config

    def download(self, download_requests, max_threads=None):
        """ Download one or multiple requests, provided as a request list.

        :param download_requests: A list of requests or a single request to be executed.
        :type download_requests: List[DownloadRequest] or DownloadRequest
        :param max_threads: Maximum number of threads to be used for the download.
        :type max_threads: int
        :return: A list of results or a single result, depending on input parameter `requests`
        :rtype: list(object) or object
        """
        is_single_request = isinstance(download_requests, DownloadRequest)
        if is_single_request:
            download_requests = [download_requests]

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
            download_list = [executor.submit(self._single_download, request) for request in download_requests]

        data_list = []
        for future in download_list:
            try:
                data_list.append(future.result())
            except DownloadFailedException as download_exception:
                if self.raise_download_errors:
                    traceback = sys.exc_info()[2]
                    raise download_exception.with_traceback(traceback)

                warnings.warn(str(download_exception), category=SHRuntimeWarning)
                data_list.append(None)

        if is_single_request:
            return data_list[0]
        return data_list

    def _single_download(self, request):
        """ Method for downloading a single request

        :param request: An object with information about download and storage of data
        :type request: DownloadRequest
        :return: Downloaded data
        :rtype: object
        """
        request.raise_if_invalid()

        if request.save_response:
            request.file_path = os.path.join(request.data_folder, request.file_path)
        elif request.hash_save:
            hashed, hashable = hash_request(request.url, request.post_values)
            folder_path = os.path.join(request.data_folder, hashed)
            request.file_path = os.path.join(folder_path, 'response.{}'.format(request.data_type.value))
            request_path = os.path.join(folder_path, 'request.json')
            LOGGER.debug('Saving hashed request to %s', request_path)
            self.io_object.write(request_path, hashable, data_format=MimeType.TXT)

        if not self.is_download_required(request):
            if request.return_data:
                return self.io_object.read(request.file_path, data_format=request.data_type)
            return None

        response_content = self._execute_download(request)

        if request.save_response or request.hash_save:
            self.io_object.write(request.file_path, response_content, data_format=MimeType.RAW)
            LOGGER.debug('Saved data to %s', request.file_path)

        if request.return_data:
            return decode_data(response_content, request.data_type)
        return None

    @retry_temporal_errors
    @fail_user_errors
    def _execute_download(self, request):
        response = requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=request.headers,
            timeout=self.config.download_timeout_seconds
        )

        response.raise_for_status()
        LOGGER.debug('Successful download from %s', request.url)

        return response.content

    def is_download_required(self, request):
        """ Checks if download should actually be done

        :param request: An object with information about download and storage of data
        :type request: DownloadRequest
        :return: True if download should be done and False otherwise
        :rtype: bool
        """
        val = (request.save_response or request.return_data) and \
              (self.redownload or request.file_path is None or not self.io_object.exists(request.file_path))
        return val


def get_json(url, post_values=None, headers=None, download_client_class=DownloadClient):
    """ Download request as JSON data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :param post_values: form encoded data to send in POST request. Default is `None`
    :type post_values: dict
    :param headers: add HTTP headers to request. Default is `None`
    :type headers: dict
    :return: request response as JSON instance
    :rtype: JSON instance or None
    :raises: RunTimeError
    """
    json_headers = {} if headers is None else headers

    if post_values is None:
        request_type = RequestType.GET
    else:
        request_type = RequestType.POST
        json_headers = {**json_headers, **{'Content-Type': MimeType.JSON.get_string()}}

    request = DownloadRequest(url=url, headers=json_headers, request_type=request_type, post_values=post_values,
                              save_response=False, return_data=True, data_type=MimeType.JSON)

    return download_client_class().download(request)


def get_xml(url, download_client_class=DownloadClient):
    """ Download request as XML data type

    :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
    :type url: str
    :return: request response as XML instance
    :rtype: XML instance or None
    :raises: RunTimeError
    """
    request = DownloadRequest(url=url, request_type=RequestType.GET, save_response=False, return_data=True,
                              data_type=MimeType.XML)

    return download_client_class().download(request)



""" Helper functions for generating request dictionaries

Request structure documentation is available at: https://docs.sentinel-hub.com/api/latest/reference/
"""

from .geometry import Geometry


def body(request_bounds, request_data, evalscript, request_output=None):
    """ Generate request body
    """
    request_body = {
        "input": {
            "bounds": request_bounds,
            "data": request_data
        },
        "evalscript": evalscript
    }

    if request_output is not None:
        request_body['output'] = request_output

    return request_body


def response(identifier, response_format):
    """ Generate request response
    """
    return {
        "identifier": identifier,
        "format": {
            'type': response_format
        }
    }


def output(responses, size_x, size_y):
    """ Generate request output
    """
    return {
        "width": size_x,
        "height": size_y,
        "responses": responses
    }


def bounds(crs, bbox=None, geometry=None):
    """ Generate request bounds
    """
    if bbox is None and geometry is None:
        raise ValueError("At least one of parameters 'bbox' and 'geometry' has to be given")

    if bbox and (not isinstance(bbox, list) or len(bbox) != 4 or not all(isinstance(x, float) for x in bbox)):
        raise ValueError("Invalid bbox argument: {}".format(bbox))

    if geometry and not isinstance(geometry, Geometry):
        raise ValueError('Geometry has to be of type sentinelhub.Geometry')

    if bbox and geometry and bbox is not geometry.crs:
        raise ValueError('Bounding box and geometry should have the same CRS, but {} and {} '
                         'found'.format(bbox, geometry.crs))

    request_bounds = {
        "properties": {
            "crs": crs
        }
    }

    if bbox:
        request_bounds['bbox'] = list(bbox)

    if geometry:
        request_bounds['geometry'] = geometry.geojson

    return request_bounds


def data(time_from=None, time_to=None, data_type='S2L1C'):
    """ Generate request data
    """
    return {
        "type": data_type,
        "dataFilter": {
            "timeRange": {
                "from": "" if time_from is None else time_from,
                "to": "" if time_to is None else time_to
            }
        }
    }



"""
Utility tools for writing unit tests for packages which rely on `sentinelhub-py`
"""

import unittest
import os
import shutil
import logging
import inspect

import numpy as np

from .config import SHConfig


class TestSentinelHub(unittest.TestCase):
    """ Class implementing common functionalities of unit tests for working with `sentinelhub-py` package:

    - reading configuration parameters from environmental variables and saving them to `config.json`,
    - setting logger,
    - handling input and output data folders,
    - method for testing statistics of a numpy data array.
    """

    CONFIG = None
    LOGGER = None
    INPUT_FOLDER = None
    OUTPUT_FOLDER = None

    INPUT_FOLDER = None
    OUTPUT_FOLDER = None
    CLEAR_OUTPUTS = True

    @classmethod
    def setUpClass(cls):
        """ A general set up class

        Use ``super().setUpClass()`` in every class which inherits ``TestSentinelHub``
        """
        if cls.__name__ == TestSentinelHub.__name__:
            return

        cls.INPUT_FOLDER = os.path.join(os.path.dirname(inspect.getsourcefile(cls)), 'TestInputs')
        cls.OUTPUT_FOLDER = os.path.join(os.path.dirname(inspect.getsourcefile(cls)), 'TestOutputs')

        if cls.CONFIG is None:
            cls.CONFIG = cls._config_with_environment_variables()

        if cls.LOGGER is None:
            logging.basicConfig(level=logging.INFO,
                                format='%(asctime)-15s %(module)s:%(lineno)d [%(levelname)s] %(funcName)s  %(message)s')
            cls.LOGGER = logging.getLogger(__name__)

    @staticmethod
    def _config_with_environment_variables():
        """ Reads configuration parameters from environmental variables
        """
        config = SHConfig()
        for param in config.get_params():
            env_variable = param.upper()
            if os.environ.get(env_variable):
                setattr(config, param, os.environ.get(env_variable))
        config.save()
        return config

    @classmethod
    def tearDownClass(cls):
        if cls.CLEAR_OUTPUTS and cls.OUTPUT_FOLDER:
            shutil.rmtree(cls.OUTPUT_FOLDER, ignore_errors=True)

    def test_numpy_data(self, data=None, exp_shape=None, exp_dtype=None, exp_min=None, exp_max=None, exp_mean=None,
                        exp_median=None, delta=None, test_name=''):
        """ Validates basic statistics of data array

        :param data: Data array
        :type data: numpy.ndarray
        :param exp_shape: Expected shape
        :type exp_shape: tuple(int)
        :param exp_dtype: Expected dtype
        :type exp_dtype: numpy.dtype
        :param exp_min: Expected minimal value
        :type exp_min: float
        :param exp_max: Expected maximal value
        :type exp_max: float
        :param exp_mean: Expected mean value
        :type exp_mean: float
        :param exp_median: Expected median value
        :type exp_median: float
        :param delta: Precision of validation. If not set, it will be set automatically
        :type delta: float
        :param test_name: Name of the test case
        :type test_name: str
        """
        if data is None:
            return
        if delta is None:
            delta = 1e-1 if np.issubdtype(data.dtype, np.integer) else 1e-4

        for exp_stat, stat_val, stat_name in [(exp_shape, data.shape, 'shape'), (exp_dtype, data.dtype, 'dtype')]:
            if exp_stat is not None:
                with self.subTest(msg='Test case {}'.format(test_name)):
                    self.assertEqual(stat_val, exp_stat,
                                     msg='Expected {} {}, got {}'.format(stat_name, exp_stat, stat_val))

        for exp_stat, stat_func, stat_name in [(exp_min, np.amin, 'min'), (exp_max, np.amax, 'max'),
                                               (exp_mean, np.mean, 'mean'), (exp_median, np.median, 'median')]:
            if exp_stat is not None:
                stat_val = stat_func(data)
                with self.subTest(msg='Test case {}'.format(test_name)):
                    self.assertAlmostEqual(stat_val, exp_stat, delta=delta,
                                           msg='Expected {} {}, got {}'.format(stat_name, exp_stat, stat_val))


class TestCaseContainer:
    """ Class for storing expected statistics for a single test case

    :param name: Name of a test case
    :type name: str
    :param request: A class which provides the data for testing
    :type request: object
    :param stats: Any other parameters
    """
    def __init__(self, name, request, **stats):
        self.name = name
        self.request = request

        for stat_name, stat_value in stats.items():
            setattr(self, stat_name, stat_value)

    def __getattr__(self, key):
        """ Fallback if the attribute is missing - in that case `None` is returned
        """



"""
Module implementing Sentinel Hub session object
"""
import time

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from .config import SHConfig


class SentinelHubSession:
    """ Sentinel Hub authentication class

    The class will do OAuth2 authentication with Sentinel Hub service and store the token. It will make sure that the
    token is never expired by automatically refreshing it if expiry time is close.
    """
    SECONDS_BEFORE_EXPIRY = 60

    def __init__(self, config=None):
        """
        :param config: An instance of package configuration class
        :type config: SHConfig
        """
        self.config = SHConfig() if config is None else config

        if not (self.config.sh_client_id and self.config.sh_client_secret):
            raise ValueError("Configuration parameters 'sh_client_id' and 'sh_client_secret' have to be set in order"
                             "to authenticate with Sentinel Hub service")

        self._token = None

    @property
    def token(self):
        """ Always up-to-date session's token

        :return: A token in a form of dictionary of parameters
        :rtype: dict
        """
        if self._token and self._token['expires_at'] > time.time() + self.SECONDS_BEFORE_EXPIRY:
            return self._token

        self._token = self._fetch_token()
        return self._token

    @property
    def session_headers(self):
        """ Provides session authorization headers

        :return: A dictionary with authorization headers
        :rtype: dict
        """
        return {
            'Authorization': 'Bearer {}'.format(self.token['access_token'])
        }

    def _fetch_token(self):
        """ Collects a new token from Sentinel Hub service
        """
        oauth_client = BackendApplicationClient(client_id=self.config.sh_client_id)

        with OAuth2Session(client=oauth_client) as oauth_session:
            return oauth_session.fetch_token(
                token_url=self.config.get_sh_oauth_url(),
                client_id=self.config.sh_client_id,
                client_secret=self.config.sh_client_secret
            )



"""
Module defining constants and enumerate types used in the package
"""
import re
import functools
import itertools as it
import mimetypes
from enum import Enum, EnumMeta
from aenum import extend_enum

import utm
import pyproj

from .config import SHConfig
from ._version import __version__


class PackageProps:
    """ Class for obtaining package properties. Currently it supports obtaining package version."""

    @staticmethod
    def get_version():
        """ Returns package version

        :return: package version
        :rtype: str
        """
        return __version__


class ServiceType(Enum):
    """ Enum constant class for type of service

    Supported types are WMS, WCS, WFS, AWS, IMAGE
    """
    WMS = 'wms'
    WCS = 'wcs'
    WFS = 'wfs'
    AWS = 'aws'
    IMAGE = 'image'
    FIS = 'fis'


class _Source(Enum):
    """
    Types of satellite sources
    """
    SENTINEL2 = 'Sentinel-2'
    SENTINEL1 = 'Sentinel-1'
    LANDSAT8 = 'Landsat 8'
    MODIS = 'MODIS'
    DEM = 'Mapzen DEM'
    LANDSAT5 = 'Landsat 5'
    LANDSAT7 = 'Landsat 7'
    SENTINEL3 = 'Sentinel-3'
    SENTINEL5P = 'Sentinel-5P'
    ENVISAT_MERIS = 'Envisat Meris'


class _ProcessingLevel(Enum):
    """
    Types of processing level
    """
    # pylint: disable=invalid-name
    L2 = 'L2'
    L1C = 'L1C'
    L2A = 'L2A'
    L3B = 'L3B'
    L1TP = 'L1TP'
    GRD = 'GRD'
    MCD43A4 = 'MCD43A4'


class _Acquisition(Enum):
    """
    Types of satellite acquisition
    """
    # pylint: disable=invalid-name
    IW = 'IW'
    EW = 'EW'
    OLCI = 'OLCI'


class _Polarisation(Enum):
    """
    Types of Sentinel-1 polarisation
    """
    # pylint: disable=invalid-name
    DV = 'VV+VH'
    DH = 'HH+HV'
    SV = 'VV'
    SH = 'HH'


class _Resolution(Enum):
    """
    Types of Sentinel-1 resolution
    """
    MEDIUM = 'medium'
    HIGH = 'high'


class _OrbitDirection(Enum):
    """
    Types of Sentinel-1 orbit direction
    """
    ASCENDING = 'ascending'
    DESCENDING = 'descending'
    BOTH = 'both'


class DataSourceMeta(EnumMeta):
    """
    Meta class used to add custom (BYOC) datasources to DataSource enumerator.

    BYOC stands for Bring Your Own Data. For more details see: https://docs.sentinel-hub.com/api/latest/#/API/byoc
    """
    def __call__(cls, collection_id, *args, **kwargs):
        """
        This is executed whenever DataSource('something') is called.

        The method raises a ValueError if the 'something' does not match the format expected for collection id.
        """
        if not isinstance(collection_id, str):
            return super().__call__(collection_id, *args, **kwargs)

        collection_id_pattern = '.{8}-.{4}-.{4}-.{4}-.{12}'
        if not re.compile(collection_id_pattern).match(collection_id):
            raise ValueError("Given collection id does not match the expected format {}".format(collection_id_pattern))

        datasource_name = cls._custom_datasource_name(collection_id)
        if datasource_name not in cls._member_names_:
            extend_enum(cls, datasource_name, collection_id)

        return super().__call__(collection_id, *args, **kwargs)

    @staticmethod
    def _custom_datasource_name(collection_id):
        """
        Prepares a name for custom (BYOC) datasource, which is then added to DataSource enum.

        :param: collection_id: Collection id of the BYOC, user's input.
        :type: string
        :return: Name for custom (BYOC) datasource.
        :rtype: string
        """
        return 'BYOC_{}'.format(collection_id)


class DataSource(Enum, metaclass=DataSourceMeta):
    """ Enum constant class for types of satellite data

    Supported types are SENTINEL2_L1C, SENTINEL2_L2A, LANDSAT8, SENTINEL1_IW, SENTINEL1_EW, SENTINEL1_EW_SH,
    SENTINEL1_IW_ASC, SENTINEL1_EW_ASC, SENTINEL1_EW_SH_ASC, SENTINEL1_IW_DES, SENTINEL1_EW_DES, SENTINEL1_EW_SH_DES,
    DEM, MODIS, LANDSAT5, LANDSAT7, SENTINEL3, SENTINEL5P, ENVISAT_MERIS, SENTINEL2_L3B, LANDSAT8_L2A
    """
    SENTINEL2_L1C = (_Source.SENTINEL2, _ProcessingLevel.L1C)
    SENTINEL2_L2A = (_Source.SENTINEL2, _ProcessingLevel.L2A)
    SENTINEL1_IW = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.IW, _Polarisation.DV, _Resolution.HIGH,
                    _OrbitDirection.BOTH)
    SENTINEL1_EW = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.DH, _Resolution.MEDIUM,
                    _OrbitDirection.BOTH)
    SENTINEL1_EW_SH = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.SH, _Resolution.MEDIUM,
                       _OrbitDirection.BOTH)
    SENTINEL1_IW_ASC = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.IW, _Polarisation.DV, _Resolution.HIGH,
                        _OrbitDirection.ASCENDING)
    SENTINEL1_EW_ASC = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.DH, _Resolution.MEDIUM,
                        _OrbitDirection.ASCENDING)
    SENTINEL1_EW_SH_ASC = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.SH,
                           _Resolution.MEDIUM, _OrbitDirection.ASCENDING)
    SENTINEL1_IW_DES = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.IW, _Polarisation.DV, _Resolution.HIGH,
                        _OrbitDirection.DESCENDING)
    SENTINEL1_EW_DES = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.DH, _Resolution.MEDIUM,
                        _OrbitDirection.DESCENDING)
    SENTINEL1_EW_SH_DES = (_Source.SENTINEL1, _ProcessingLevel.GRD, _Acquisition.EW, _Polarisation.SH,
                           _Resolution.MEDIUM, _OrbitDirection.DESCENDING)
    DEM = (_Source.DEM, )
    MODIS = (_Source.MODIS, _ProcessingLevel.MCD43A4)
    LANDSAT8 = (_Source.LANDSAT8, _ProcessingLevel.L1TP)
    #: custom BYOC enum members are defined in DataSourceMeta
    # eocloud sources:
    LANDSAT5 = (_Source.LANDSAT5, _ProcessingLevel.GRD)
    LANDSAT7 = (_Source.LANDSAT7, _ProcessingLevel.GRD)
    SENTINEL3 = (_Source.SENTINEL3, _ProcessingLevel.L2, _Acquisition.OLCI)
    SENTINEL5P = (_Source.SENTINEL5P, _ProcessingLevel.L2)
    ENVISAT_MERIS = (_Source.ENVISAT_MERIS, )
    SENTINEL2_L3B = (_Source.SENTINEL2, _ProcessingLevel.L3B)
    LANDSAT8_L2A = (_Source.LANDSAT8, _ProcessingLevel.L2A)
    LANDSAT8_L1C = (_Source.LANDSAT8, _ProcessingLevel.L1C)

    @classmethod
    def get_wfs_typename(cls, data_source):
        """ Maps data source to string identifier for WFS

        :param data_source: One of the supported data sources
        :type data_source: DataSource
        :return: Product identifier for WFS
        :rtype: str
        """
        is_eocloud = SHConfig().has_eocloud_url()

        if data_source.is_custom():
            return 'DSS10-{}'.format(data_source.value)

        return {
            cls.SENTINEL2_L1C: 'S2.TILE',
            cls.SENTINEL2_L2A: 'SEN4CAP_S2L2A.TILE' if is_eocloud else 'DSS2',
            cls.SENTINEL1_IW: 'S1.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW: 'S1_EW.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW_SH: 'S1_EW_SH.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_IW_ASC: 'S1.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW_ASC: 'S1_EW.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW_SH_ASC: 'S1_EW_SH.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_IW_DES: 'S1.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW_DES: 'S1_EW.TILE' if is_eocloud else 'DSS3',
            cls.SENTINEL1_EW_SH_DES: 'S1_EW_SH.TILE' if is_eocloud else 'DSS3',
            cls.DEM: 'DSS4',
            cls.MODIS: 'DSS5',
            cls.LANDSAT8: 'L8.TILE' if is_eocloud else 'DSS6',
            # eocloud sources only:
            cls.LANDSAT5: 'L5.TILE',
            cls.LANDSAT7: 'L7.TILE',
            cls.SENTINEL3: 'S3.TILE',
            cls.SENTINEL5P: 'S5p_L2.TILE',
            cls.ENVISAT_MERIS: 'ENV.TILE',
            cls.SENTINEL2_L3B: 'SEN4CAP_S2L3B.TILE',
            cls.LANDSAT8_L2A: 'SEN4CAP_L8L2A.TILE'
        }[data_source]

    def api_identifier(self):
        """ Returns Sentinel Hub API identifier string

        :return: A data source identifier string
        :rtype: str
        """
        return {
            DataSource.SENTINEL2_L1C: 'S2L1C',
            DataSource.SENTINEL2_L2A: 'S2L2A',
            DataSource.LANDSAT8_L1C: 'L8L1C',
            DataSource.DEM: 'DEM',
            DataSource.MODIS: 'MODIS'
        }[self]

    def bands(self):
        """ Get available bands for a particular data source
        """
        return {
            DataSource.SENTINEL2_L1C: [
                "B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B8A", "B09", "B10", "B11", "B12"
            ],
            DataSource.SENTINEL2_L2A: [
                "B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B8A", "B09", "B11", "B12"
            ],
            DataSource.LANDSAT8_L1C: [
                "B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B09", "B10", "B11"
            ],
            DataSource.DEM: [
                "DEM"
            ],
            DataSource.MODIS: [
                "B01", "B02", "B03", "B04", "B05", "B06", "B07"
            ]
        }[self]

    def is_sentinel1(self):
        """Checks if source is Sentinel-1

        Example: ``DataSource.SENTINEL1_IW.is_sentinel1()`` or ``DataSource.is_sentinel1(DataSource.SENTINEL1_IW)``

        :param self: One of the supported data sources
        :type self: DataSource
        :return: `True` if source is Sentinel-1 and `False` otherwise
        :rtype: bool
        """
        return self.value[0] is _Source.SENTINEL1

    def contains_orbit_direction(self, orbit_direction):
        """Checks if Sentine-1 data source contains given orbit direction.
        Note: Data sources with "both" orbit directions contain ascending and descending orbit directions.

        :param self: One of the supported Sentinel-1 data sources
        :type self: DataSource
        :param orbit_direction: One of the orbit directions
        :type orbit_direction: string
        :return: `True` if data source contains the orbit direction
        :return: bool
        """
        if not self.is_sentinel1():
            raise ValueError("Orbit direction can only be checked for Sentinel-1 data source.")
        return self.value[5].name.upper() in [orbit_direction.upper(), _OrbitDirection.BOTH.value.upper()]

    def is_timeless(self):
        """Checks if data source is time independent

        Example: ``DataSource.DEM.is_timeless()`` or ``DataSource.is_timeless(DataSource.DEM)``

        :param self: One of the supported data sources
        :type self: DataSource
        :return: `True` if data source is time independent and `False` otherwise
        :rtype: bool
        """
        return self.value[0] is _Source.DEM

    def is_uswest_source(self):
        """Checks if data source via Sentinel Hub services is available at US West server

        Example: ``DataSource.LANDSAT8.is_uswest_source()`` or ``DataSource.is_uswest_source(DataSource.LANDSAT8)``

        :param self: One of the supported data sources
        :type self: DataSource
        :return: `True` if data source exists at US West server and `False` otherwise
        :rtype: bool
        """
        return not SHConfig().has_eocloud_url() and self.value[0] in [_Source.LANDSAT8, _Source.MODIS, _Source.DEM]

    @classmethod
    def get_available_sources(cls):
        """ Returns which data sources are available for configured Sentinel Hub OGC URL

        :return: List of available data sources
        :rtype: list(sentinelhub.DataSource)
        """
        if SHConfig().has_eocloud_url():
            return [cls.SENTINEL2_L1C, cls.SENTINEL2_L2A, cls.SENTINEL2_L3B, cls.SENTINEL1_IW, cls.SENTINEL1_EW,
                    cls.SENTINEL1_EW_SH, cls.SENTINEL3, cls.SENTINEL5P, cls.LANDSAT5, cls.LANDSAT7, cls.LANDSAT8,
                    cls.LANDSAT8_L2A, cls.ENVISAT_MERIS]

        return [cls.SENTINEL2_L1C, cls.SENTINEL2_L2A, cls.SENTINEL1_IW, cls.SENTINEL1_EW, cls.SENTINEL1_EW_SH,
                cls.SENTINEL1_IW_ASC, cls.SENTINEL1_EW_ASC, cls.SENTINEL1_EW_SH_ASC, cls.SENTINEL1_IW_DES,
                cls.SENTINEL1_EW_DES, cls.SENTINEL1_EW_SH_DES, cls.DEM, cls.MODIS, cls.LANDSAT8,
                *cls.get_custom_sources()]

    @classmethod
    def get_custom_sources(cls):
        """Returns the list of all custom (BYOC) datasources, which have ben added to the DataSource enumerator.

        :return: List of custom (BYOC) datasources
        :rtype: list(sentinelhub.DataSource)
        """
        return [datasource for datasource in cls if datasource.is_custom()]

    def is_custom(self):
        """ Checks is if datasource is a custom Sentinel Hub BYOC data source

        :return: True if datasource is custom and False otherwise
        :rtype: bool
        """
        return self.name.startswith('BYOC_')


class CRSMeta(EnumMeta):
    """ Metaclass used for building CRS Enum class
    """
    def __new__(mcs, cls, bases, classdict):
        """ This is executed at the beginning of runtime when CRS class is created
        """
        for direction, direction_value in [('N', '6'), ('S', '7')]:
            for zone in range(1, 61):
                classdict['UTM_{}{}'.format(zone, direction)] = '32{}{}'.format(direction_value, str(zone).zfill(2))

        return super().__new__(mcs, cls, bases, classdict)

    def __call__(cls, value, *args, **kwargs):
        """ This is executed whenever CRS('something') is called
        """
        return super().__call__(CRSMeta._parse_crs(value), *args, **kwargs)

    @staticmethod
    def _parse_crs(value):
        """ Method for parsing different inputs representing the same CRS enum. Example:
        """
        if isinstance(value, int):
            return str(value)
        if isinstance(value, str):
            return value.strip('epsgEPSG: ')
        return value


class CRS(Enum, metaclass=CRSMeta):
    """ Coordinate Reference System enumerate class

    Available CRS constants are WGS84, POP_WEB (i.e. Popular Web Mercator) and constants in form UTM_<zone><direction>,
    where zone is an integer from [1, 60] and direction is either N or S (i.e. northern or southern hemisphere)
    """
    WGS84 = '4326'
    POP_WEB = '3857'
    #: UTM enum members are defined in CRSMeta.__new__

    def __str__(self):
        """ Method for casting CRS enum into string
        """
        return self.ogc_string()

    @classmethod
    def has_value(cls, value):
        """ Tests whether CRS contains a constant defined with string `value`.

        :param value: The string representation of the enum constant.
        :type value: str
        :return: `True` if there exists a constant with string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)

    @property
    def epsg(self):
        """ EPSG code property

        :return: EPSG code of given CRS
        :rtype: int
        """
        return int(self.value)

    def ogc_string(self):
        """ Returns a string of the form authority:id representing the CRS.

        :param self: An enum constant representing a coordinate reference system.
        :type self: CRS
        :return: A string representation of the CRS.
        :rtype: str
        """
        return 'EPSG:{}'.format(CRS(self).value)

    @property
    def opengis_string(self):
        """ Returns an URL to OGC webpage where the CRS is defined

        :return: An URL with CRS definition
        :rtype: str
        """
        return 'http://www.opengis.net/def/crs/EPSG/0/{}'.format(self.epsg)

    def is_utm(self):
        """ Checks if crs is one of the 64 possible UTM coordinate reference systems.

        :param self: An enum constant representing a coordinate reference system.
        :type self: CRS
        :return: `True` if crs is UTM and `False` otherwise
        :rtype: bool
        """
        return self.name.startswith('UTM')

    @functools.lru_cache(maxsize=5)
    def projection(self):
        """ Returns a projection in form of pyproj class. For better time performance it will cache results of
        5 most recently used CRS classes.

        :param self: An enum constant representing a coordinate reference system.
        :type self: CRS
        :return: pyproj projection class
        :rtype: pyproj.Proj
        """
        return pyproj.Proj(init=self.ogc_string(), preserve_units=True)

    @functools.lru_cache(maxsize=10)
    def get_transform_function(self, other):
        """ Returns a function for transforming geometrical objects from one CRS to another. The function will support
        transformations between any objects that pyproj supports.
        For better time performance this method will cache results of 10 most recently used pairs of CRS classes.

        :param self: Initial CRS
        :type self: CRS
        :param other: Target CRS
        :type other: CRS
        :return: A projection function obtained from pyproj package
        :rtype: function
        """
        if pyproj.__version__ >= '2':
            return pyproj.Transformer.from_proj(self.projection(), other.projection(), skip_equivalent=True).transform

        return functools.partial(pyproj.transform, self.projection(), other.projection())

    @staticmethod
    def get_utm_from_wgs84(lng, lat):
        """ Convert from WGS84 to UTM coordinate system

        :param lng: Longitude
        :type lng: float
        :param lat: Latitude
        :type lat: float
        :return: UTM coordinates
        :rtype: tuple
        """
        _, _, zone, _ = utm.from_latlon(lat, lng)
        direction = 'N' if lat >= 0 else 'S'
        return CRS['UTM_{}{}'.format(str(zone), direction)]


class CustomUrlParam(Enum):
    """ Enum class to represent supported custom url parameters of OGC services

    Supported parameters are `SHOWLOGO`, `ATMFILTER`, `EVALSCRIPT`, `EVALSCRIPTURL`, `PREVIEW`, `QUALITY`, `UPSAMPLING`,
    `DOWNSAMPLING`, `TRANSPARENT`, `BGCOLOR` and `GEOMETRY`.

    See http://sentinel-hub.com/develop/documentation/api/custom-url-parameters and
    https://www.sentinel-hub.com/develop/documentation/api/ogc_api/wms-parameters for more information.
    """
    SHOWLOGO = 'ShowLogo'
    ATMFILTER = 'AtmFilter'
    EVALSCRIPT = 'EvalScript'
    EVALSCRIPTURL = 'EvalScriptUrl'
    PREVIEW = 'Preview'
    QUALITY = 'Quality'
    UPSAMPLING = 'Upsampling'
    DOWNSAMPLING = 'Downsampling'
    TRANSPARENT = 'Transparent'
    BGCOLOR = 'BgColor'
    GEOMETRY = 'Geometry'

    @classmethod
    def has_value(cls, value):
        """ Tests whether CustomUrlParam contains a constant defined with a string `value`

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with a string value `value`, `False` otherwise
        :rtype: bool
        """
        return any(value.lower() == item.value.lower() for item in cls)

    @staticmethod
    def get_string(param):
        """ Get custom url parameter name as string

        :param param: CustomUrlParam enum constant
        :type param: Enum constant
        :return: String describing the file format
        :rtype: str
        """
        return param.value


class HistogramType(Enum):
    """ Enum class for types of histogram supported by Sentinel Hub FIS service

    Supported histogram types are EQUALFREQUENCY, EQUIDISTANT and STREAMING
    """
    EQUALFREQUENCY = 'equalfrequency'
    EQUIDISTANT = 'equidistant'
    STREAMING = 'streaming'


class MimeType(Enum):
    """ Enum class to represent supported image file formats

    Supported file formats are TIFF 8-bit, TIFF 16-bit, TIFF 32-bit float, PNG, JPEG, JPEG2000, JSON, CSV, ZIP, HDF5,
    XML, GML, RAW
    """
    TIFF = 'tiff'
    TIFF_d8 = 'tiff;depth=8'
    TIFF_d16 = 'tiff;depth=16'  # This is the same as TIFF
    TIFF_d32f = 'tiff;depth=32f'
    PNG = 'png'
    JPG = 'jpg'
    JP2 = 'jp2'
    JSON = 'json'
    CSV = 'csv'
    ZIP = 'zip'
    HDF = 'hdf'
    XML = 'xml'
    GML = 'gml'
    TXT = 'txt'
    TAR = 'tar'
    RAW = 'raw'
    SAFE = 'safe'

    @staticmethod
    def canonical_extension(fmt_ext):
        """ Canonical extension of file format extension

        Converts the format extension fmt_ext into the canonical extension for that format. For example,
        ``canonical_extension('tif') == 'tiff'``. Here we agree that the canonical extension for format F is F.value

        :param fmt_ext: A string representing an extension (e.g. ``'txt'``, ``'png'``, etc.)
        :type fmt_ext: str
        :return: The canonical form of the extension (e.g. if ``fmt_ext='tif'`` then we return ``'tiff'``)
        :rtype: str
        """
        if MimeType.has_value(fmt_ext):
            return fmt_ext
        try:
            return {
                'tif': MimeType.TIFF.value,
                'jpeg': MimeType.JPG.value,
                'hdf5': MimeType.HDF.value,
                'h5': MimeType.HDF.value
            }[fmt_ext]
        except KeyError:
            raise ValueError('Data format .{} is not supported'.format(fmt_ext))

    def is_image_format(self):
        """ Checks whether file format is an image format

        Example: ``MimeType.PNG.is_image_format()`` or ``MimeType.is_image_format(MimeType.PNG)``

        :param self: File format
        :type self: MimeType
        :return: `True` if file is in image format, `False` otherwise
        :rtype: bool
        """
        return self in frozenset([MimeType.TIFF, MimeType.TIFF_d8, MimeType.TIFF_d16, MimeType.TIFF_d32f, MimeType.PNG,
                                  MimeType.JP2, MimeType.JPG])

    def is_api_format(self):
        """ Checks if mime type is supported by Sentinel Hub API

        :return: True if API supports this format and False otherwise
        :rtype: bool
        """
        return self in frozenset([MimeType.JPG, MimeType.PNG, MimeType.TIFF, MimeType.JSON])

    def is_tiff_format(self):
        """ Checks whether file format is a TIFF image format

        Example: ``MimeType.TIFF.is_tiff_format()`` or ``MimeType.is_tiff_format(MimeType.TIFF)``

        :param self: File format
        :type self: MimeType
        :return: `True` if file is in image format, `False` otherwise
        :rtype: bool
        """
        return self in frozenset([MimeType.TIFF, MimeType.TIFF_d8, MimeType.TIFF_d16, MimeType.TIFF_d32f])

    @classmethod
    def has_value(cls, value):
        """ Tests whether MimeType contains a constant defined with string ``value``

        :param value: The string representation of the enum constant
        :type value: str
        :return: `True` if there exists a constant with string value ``value``, `False` otherwise
        :rtype: bool
        """
        return any(value == item.value for item in cls)

    def get_string(self):
        """ Get file format as string

        :return: String describing the file format
        :rtype: str
        """
        if self is MimeType.TAR:
            return 'application/x-tar'
        if self is MimeType.JSON:
            return 'application/json'
        if self in [MimeType.TIFF_d8, MimeType.TIFF_d16, MimeType.TIFF_d32f]:
            return 'image/{}'.format(self.value)
        if self is MimeType.JP2:
            return 'image/jpeg2000'
        if self is MimeType.RAW:
            return self.value
        return mimetypes.types_map['.' + self.value]

    def get_sample_type(self):
        """ Returns sampleType used in Sentinel-Hub evalscripts.

        :return: sampleType
        :rtype: str
        :raises: ValueError
        """
        try:
            return {
                MimeType.TIFF: 'INT16',
                MimeType.TIFF_d8: 'INT8',
                MimeType.TIFF_d16: 'INT16',
                MimeType.TIFF_d32f: 'FLOAT32'
            }[self]
        except IndexError:
            raise ValueError('Type {} is not supported by this method'.format(self))

    def get_expected_max_value(self):
        """ Returns max value of image `MimeType` format and raises an error if it is not an image format

        Note: For `MimeType.TIFF_d32f` it will return ``1.0`` as that is expected maximum for an image even though it
        could be higher.

        :return: A maximum value of specified image format
        :rtype: int or float
        :raises: ValueError
        """
        try:
            return {
                MimeType.TIFF: 65535,
                MimeType.TIFF_d8: 255,
                MimeType.TIFF_d16: 65535,
                MimeType.TIFF_d32f: 1.0,
                MimeType.PNG: 255,
                MimeType.JPG: 255,
                MimeType.JP2: 10000
            }[self]
        except IndexError:
            raise ValueError('Type {} is not supported by this method'.format(self))

    @staticmethod
    def from_string(mime_type_str):
        """ Parses mime type from a file extension string

        :param mime_type_str: A file extension string
        :type mime_type_str: str
        :return: A mime type enum
        :rtype: MimeType
        """
        if mime_type_str == 'jpeg':
            return MimeType.JPG

        return MimeType(mime_type_str)


class RequestType(Enum):
    """ Enum constant class for GET/POST request type """
    GET = 'GET'
    POST = 'POST'


class SHConstants:
    """ Initialisation of constants used by OGC request.

        Constants are LATEST
    """
    LATEST = 'latest'
    HEADERS = {'User-Agent': 'sentinelhub-py/v{}'.format(PackageProps.get_version())}


class AwsConstants:
    """ Initialisation of every constant used by AWS classes

    For each supported data source it contains lists of all possible bands and all posible metadata files:

        - S2_L1C_BANDS and S2_L1C_METAFILES
        - S2_L2A_BANDS and S2_L2A_METAFILES

    It also contains dictionary of all possible files and their formats: AWS_FILES
    """
    # General constants:
    SOURCE_ID_LIST = ['L1C', 'L2A']
    TILE_INFO = 'tileInfo'
    PRODUCT_INFO = 'productInfo'
    METADATA = 'metadata'
    PREVIEW = 'preview'
    PREVIEW_JP2 = 'preview*'
    QI_LIST = ['DEFECT', 'DETFOO', 'NODATA', 'SATURA', 'TECQUA']
    QI_MSK_CLOUD = 'qi/MSK_CLOUDS_B00'
    AUX_DATA = 'AUX_DATA'
    DATASTRIP = 'DATASTRIP'
    GRANULE = 'GRANULE'
    HTML = 'HTML'
    INFO = 'rep_info'
    QI_DATA = 'QI_DATA'
    IMG_DATA = 'IMG_DATA'
    INSPIRE = 'inspire'
    MANIFEST = 'manifest'
    TCI = 'TCI'
    PVI = 'PVI'
    ECMWFT = 'auxiliary/ECMWFT'
    AUX_ECMWFT = 'auxiliary/AUX_ECMWFT'
    DATASTRIP_METADATA = 'datastrip/*/metadata'

    # More constants about L2A
    AOT = 'AOT'
    WVP = 'WVP'
    SCL = 'SCL'
    VIS = 'VIS'
    L2A_MANIFEST = 'L2AManifest'
    REPORT = 'report'
    GIPP = 'auxiliary/GIP_TL'
    FORMAT_CORRECTNESS = 'FORMAT_CORRECTNESS'
    GENERAL_QUALITY = 'GENERAL_QUALITY'
    GEOMETRIC_QUALITY = 'GEOMETRIC_QUALITY'
    RADIOMETRIC_QUALITY = 'RADIOMETRIC_QUALITY'
    SENSOR_QUALITY = 'SENSOR_QUALITY'
    QUALITY_REPORTS = [FORMAT_CORRECTNESS, GENERAL_QUALITY, GEOMETRIC_QUALITY, RADIOMETRIC_QUALITY, SENSOR_QUALITY]
    CLASS_MASKS = ['SNW', 'CLD']
    R10m = 'R10m'
    R20m = 'R20m'
    R60m = 'R60m'
    RESOLUTIONS = [R10m, R20m, R60m]
    S2_L2A_BAND_MAP = {R10m: ['B02', 'B03', 'B04', 'B08', AOT, TCI, WVP],
                       R20m: ['B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B11', 'B12', AOT, SCL, TCI, VIS, WVP],
                       R60m: ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B8A', 'B09', 'B11', 'B12', AOT, SCL,
                              TCI, WVP]}

    # Order of elements in following lists is important
    # Sentinel-2 L1C products:
    S2_L1C_BANDS = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
    S2_L1C_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, DATASTRIP_METADATA] +\
                       [PREVIEW, PREVIEW_JP2, TCI] +\
                       ['{}/{}'.format(preview, band) for preview, band in it.zip_longest([], S2_L1C_BANDS,
                                                                                          fillvalue=PREVIEW)] +\
                       [QI_MSK_CLOUD] +\
                       ['qi/MSK_{}_{}'.format(qi, band) for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] + \
                       ['qi/{}'.format(qi_report) for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY,
                                                                    GEOMETRIC_QUALITY, SENSOR_QUALITY]] +\
                       [ECMWFT]

    # Sentinel-2 L2A products:
    S2_L2A_BANDS = ['{}/{}'.format(resolution, band) for resolution, band_list in sorted(S2_L2A_BAND_MAP.items())
                    for band in band_list]
    S2_L2A_METAFILES = [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, L2A_MANIFEST, REPORT,
                        DATASTRIP_METADATA] + ['datastrip/*/qi/{}'.format(qi_report)for qi_report in QUALITY_REPORTS] +\
                       ['qi/{}_PVI'.format(source_id) for source_id in SOURCE_ID_LIST] +\
                       ['qi/{}_{}'.format(mask, res.lstrip('R')) for mask, res in it.product(CLASS_MASKS,
                                                                                             [R20m, R60m])] +\
                       ['qi/MSK_{}_{}'.format(qi, band) for qi, band in it.product(QI_LIST, S2_L1C_BANDS)] +\
                       [QI_MSK_CLOUD] +\
                       ['qi/{}'.format(qi_report) for qi_report in QUALITY_REPORTS] +\
                       [ECMWFT, AUX_ECMWFT, GIPP]

    # Product files with formats:
    PRODUCT_FILES = {**{PRODUCT_INFO: MimeType.JSON,
                        METADATA: MimeType.XML,
                        INSPIRE: MimeType.XML,
                        MANIFEST: MimeType.SAFE,
                        L2A_MANIFEST: MimeType.XML,
                        REPORT: MimeType.XML,
                        DATASTRIP_METADATA: MimeType.XML},
                     **{'datastrip/*/qi/{}'.format(qi_report): MimeType.XML for qi_report in QUALITY_REPORTS}}
    # Tile files with formats:
    TILE_FILES = {**{TILE_INFO: MimeType.JSON,
                     PRODUCT_INFO: MimeType.JSON,
                     METADATA: MimeType.XML,
                     PREVIEW: MimeType.JPG,
                     PREVIEW_JP2: MimeType.JP2,
                     TCI: MimeType.JP2,
                     QI_MSK_CLOUD: MimeType.GML,
                     ECMWFT: MimeType.RAW,
                     AUX_ECMWFT: MimeType.RAW,
                     GIPP: MimeType.XML},
                  **{'qi/{}'.format(qi_report): MimeType.XML for qi_report in QUALITY_REPORTS},
                  **{'{}/{}'.format(preview, band): MimeType.JP2
                     for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)},
                  **{'qi/MSK_{}_{}'.format(qi, band): MimeType.GML for qi, band in it.product(QI_LIST, S2_L1C_BANDS)},
                  **{band: MimeType.JP2 for band in S2_L1C_BANDS},
                  **{band: MimeType.JP2 for band in S2_L2A_BANDS},
                  **{'qi/{}_PVI'.format(source_id): MimeType.JP2 for source_id in SOURCE_ID_LIST},
                  **{'qi/{}_{}'.format(mask, res.lstrip('R')): MimeType.JP2 for mask, res in it.product(CLASS_MASKS,
                                                                                                        [R20m, R60m])}}

    # All files joined together
    AWS_FILES = {**PRODUCT_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in PRODUCT_FILES.items()},
                 **TILE_FILES,
                 **{filename.split('/')[-1]: data_format for filename, data_format in TILE_FILES.items()}}


class EsaSafeType(Enum):
    """ Enum constants class for ESA .SAFE type.

     Types are OLD_TYPE and COMPACT_TYPE
    """
    OLD_TYPE = 'old_type'
    COMPACT_TYPE = 'compact_type'
